"""LLM-based intelligence components for agent reasoning and action execution."""

from .config import (
    LLMInitParameters,
    LLMProfile,
    LLMStreamParameters,
    MonitoringConfig,
    OutputConfig,
    VisionConfig,
)
from .context_builder import (
    ContentFilterConfig,
    ContextBuilder,
    Frame,
    TaxonomyConfig,
    TemplateContext,
)
from .metrics_tracker import MetricsTracker, StreamMetrics
from .template_manager import TemplateManager
from .tools import TOOL_USAGE_GUIDELINES, LLMTools, ToolExecutionContext
from .vision_processor import VisionProcessor

__all__ = [
    # Config
    "LLMProfile",
    "VisionConfig",
    "MonitoringConfig",
    "OutputConfig",
    "LLMInitParameters",
    "LLMStreamParameters",
    # Components
    "ContextBuilder",
    "TemplateContext",
    "TaxonomyConfig",
    "ContentFilterConfig",
    "Frame",
    "VisionProcessor",
    "TemplateManager",
    "MetricsTracker",
    "StreamMetrics",
    # Tools
    "LLMTools",
    "ToolExecutionContext",
    "TOOL_USAGE_GUIDELINES",
]



#!/usr/bin/env python3
#
# Aiko Service: StorageFile: File-system based implementation
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Storage uses the Composite design pattern, where ...
# - A Dependency is a Component
# - A Category   is a Container and also a Dependency, i.e Component
#
# An Entry can be either a Category or Dependency, where ...
# - A Category is a directory containing Entries, referenced by a symbolic link
# - A Dependency is a file, referenced by a symbolic link
#
# Can use either "aiko_storage_file" (preferred) or "./storage_file.py"
#
# Usage: Distributed Actor
# ~~~~~~~~~~~~~~~~~~~~~~~~
# export STORAGE_RANDOM_UID=False  # debug using predictable incrementing UIDs
#
# aiko_storage_file initialize [STORAGE_URL]
# aiko_storage_file dump  [-b] [--sort_by_name]
# aiko_storage_file run        [STORAGE_URL] # file in-memory sqlite mqtt valkey
# aiko_storage_file exit
#
# aiko_storage_file add     [-b] DEPENDENCY        # file
# aiko_storage_file create  [-b] CATEGORY          # directory
# aiko_storage_file destroy [-b] ENTRY             # directory or file
# aiko_storage_file link    [-b] NEW_ENTRY EXISTING_ENTRY
# aiko_storage_file list    [-b] [-l] [-r] [PATH]
# aiko_storage_file remove  [-b] ENTRY             # directory or file
# aiko_storage_file update  [-b] service_filter lcm_url storage_url      # TODO
#
# Usage: Bootstrap [--bootstrap] or [-b]
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# All of the Distributed Actor commands also work standalone (bootstrap mode).
# These specific commands (below) are used to start from a clean slate.
#
# aiko_storage_file initialize  # set-up a file-system directory for StorageFile
# aiko_storage_file dump -b     # show low-level storage directories/files/links
#
# To Do
# ~~~~~
# * Default "_STORAGE_FILENAME" becomes "_storage_", override via "storage_url"
#   * HyperSpace "storage_url" value will become "_hyperspace_"
#
# * Implement store and load for Dependency, LifeCycleManagerURL and StorageURL
#   * LifeCycleManager, HyperSpace(LCM) and ProcessManager(LCM) use these values
#
# - Implement "aiko_storage_file repl" for interactive CRUD, etc
#
# * Validate StorageFile structure
#   * Checks "_storage_/tracked_paths" is correct (using predictable UIDs)
#   * Checks file-system ensuring everything linked together properly
#   * Checks file-system ensuring that soft-links aren't "dead"
#   * Checks that file contents are correct (using predicatable data)
#   * Checks that directory contents are correct
#   * Checks file-system against a known good "backed-up" copy
#
# * Carefully work through the semantics of add, create, destroy and remove !
#   * Consider current "_storage_/tracked_paths" versus reference counting ?
#
# * Implement "update" subcommand
# * Implement "aiko_storage_file list -c"   Category entry count
# * Implement "aiko_storage_file list -r 2" Recursion limit (default: 2)
# * Implement "aiko_storage_file list -u"   Entry UID
#
# * Implement "aiko_storage_file destroy --recursive" ... destroy everything !
#   * "rm -rf _hyperspace_ .root"
#
# * HyperSpace/StorageFile uses FileSystemEventPatternMatch in "scheme_file.py"
# * ProcessManager bootstraps (read only) by directly using Python functions
#   * Implement HyperSpace(Actor) CRUD API ... used by ProcessManager bootstrap
#     * File-system listener --> event
#
# - Consider database-based and ValKey(Redis distributed) implementation(s)
#     import sqlite3
#     self.connection = sqlite3.connect(storage_url)
#     @click.argument("storage_url", default="file://./aiko_storage.db")

from abc import abstractmethod
import click
import hashlib
import os
from pathlib import Path
import stat
import sys

from aiko_services.main import *
from aiko_services.main.storage import *
from aiko_services.main.utilities import generate, get_hostname

__all__ = ["StorageFileImpl"]

_VERSION = 0

PROTOCOL_TYPE = "storage"
ACTOR_TYPE = f"{PROTOCOL_TYPE}_file"
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{PROTOCOL_TYPE}:{_VERSION}"

_CWD_URL = "file://./"         # relative URL for the current working directory
_RESPONSE_TOPIC = f"{aiko.topic_in}"
_ROOT_FILENAME = ".root"
_STORAGE_FILENAME = "_hyperspace_"      # TODO: Default will become "_storage_"
_TRACKED_PATHNAME = f"{_STORAGE_FILENAME}/tracked_paths"
_UID_LENGTH = 12  # 12 hex digits = 6 bytes
_UID_PATHNAME = f"{_STORAGE_FILENAME}/uid_counter"

# --------------------------------------------------------------------------- #
# Storage file-system based: Implementation

class StorageFileImpl(Storage):
    @classmethod
    def create_storage(cls, storage_name, storage_url=_CWD_URL,
        register_service=True):

        tags = ["ec=true"]
        init_args = actor_args(name=storage_name, protocol=PROTOCOL, tags=tags)
        init_args["register_service"] = register_service
        init_args["storage_url"] = storage_url
        return compose_instance(StorageFileImpl, init_args)

    def __init__(self, context, storage_url=None, register_service=True):
        context.call_init(self, "Actor", context,
            register_service=register_service)

        self.share.update({                                # Inherit from Actor
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "storage_url": storage_url,
            "uid_counter": 0
        })
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            self.logger.setLevel(str(item_value).upper())

# --------------------------------------------------------------------------- #
# Storage file-system based: API implementation

    # Add Dependency (file)
    #
    # aiko_storage_file add dependency  # Is-a Category entry

    # TODO: Hyperspace: add(self, ENTRY_PATH, dependency):  # See "def list()"
    # TODO: Hyperspace:     cat, entry_name = _category_traverse(ENTRY_PATH)
    #
    # TODO: Store Dependency information ...
    #       service_filter=None, lifecycle_manager_url=None, storage_url=None

    def add(self, dependency_name, dependency=None):
        path = Path(self._create_unique_path())
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()
        try:
            Path(dependency_name).symlink_to(f"{_ROOT_FILENAME}/{path}")
            self._track_path(path)
        except FileExistsError:
            print(f'Error: "{dependency_name}" already exists', file=sys.stderr)

    # Create category (directory)
    #
    # aiko_storage_file create category  # category only

    # TODO: Store Category information ...
    #       service_filter=None, lifecycle_manager_url=None, storage_url=None

    def create(self, category_name):
        path = Path(self._create_unique_path())
        path.mkdir(parents=True, exist_ok=True)
        rel_back = self._relative_path(str(Path.cwd() / _ROOT_FILENAME), path)
        try:
            Path(path, _ROOT_FILENAME).symlink_to(rel_back)
        except FileExistsError:
            print(f'Error: internal .root link in "{path}" already exists',
                file=sys.stderr)
            return 1
        try:
            Path(category_name).symlink_to(f"{_ROOT_FILENAME}/{path}")
        except FileExistsError:
            print(f'Error: "{category_name}" already exists', file=sys.stderr)
            return 1
        except FileNotFoundError:
            print(f'Error: "{category_name}" symbolic_link target not found',
            file=sys.stderr)
            return 1
        self._track_path(path)

    # Destroy Category (directory) or Dependency (file)
    #
    # aiko_storage_file destroy entry  # Category or Dependency

    def destroy(self, entry_name):
        entry_path = Path(entry_name)
        if entry_path.is_symlink():
            absolute_target_path = entry_path.resolve()
            entry_path.unlink()
            cwd = Path.cwd()
            try:
                storage_path = cwd / _STORAGE_FILENAME
                relative_path = absolute_target_path.relative_to(storage_path)
            except ValueError:
                return

            others = [x for x in cwd.rglob("*")  \
                            if x.is_symlink() and   \
                                x.resolve() == absolute_target_path]
            if not others:
                self._untrack_path(str(absolute_target_path))
                if absolute_target_path.is_dir():
                    for child in absolute_target_path.rglob("*"):
                        if child.is_file() or child.is_symlink():
                            child.unlink()
                        else:
                            child.rmdir()
                    absolute_target_path.rmdir()
                else:
                    absolute_target_path.unlink()
                self._clean_storage(absolute_target_path.parent)
        else:
            print(
                f'Error: "{entry_name}" is not a symbolic link',
                file=sys.stderr)
            return 1

    # List storage reverse-mapped with deduplication and optional name sort
    #
    # aiko_storage_file dump --sort_by_name

    def dump(self, sort_by_name=False):
        tracked_paths = set()
        cwd = Path.cwd()
        for path in cwd.rglob("*"):
            if path.is_symlink():
                absolute_target_path = path.resolve()
                try:
                    absolute_path = cwd / _STORAGE_FILENAME
                    relative_path = absolute_target_path.relative_to(
                                        absolute_path)
                    tracked_paths.add((str(relative_path), path.name))
                except ValueError:
                    continue
        items = sorted(
                    tracked_paths, key=lambda x: x[1] if sort_by_name else x[0])
        for relative_path, name in items:
            print(f"{relative_path}  {name}")

    # Terminate Storage Actor(Service)
    #
    # aiko_storage_file exit

    def exit(self):
        aiko.process.terminate()

    # Initialize storage directory and ".root" symbolic_link
    #
    # aiko_storage_file initialize [STORAGE_URL]

    @classmethod
    def initialize(cls, storage_url=_CWD_URL):
        cwd = Path.cwd()
        storage_pathname = cwd / _STORAGE_FILENAME
        if not storage_pathname.is_dir():
            storage_pathname.mkdir(parents=True)
            print(f"Created directory {_STORAGE_FILENAME}/")

        root_pathname = cwd / _ROOT_FILENAME
        if not root_pathname.is_symlink():
            root_pathname.symlink_to(cwd)
            print(f"Created symbolic link {_ROOT_FILENAME} -->\n  {cwd}")

        uid_counter = 0
        uid_pathname = cwd / _UID_PATHNAME
        if uid_pathname.exists():
            try:
                uid_counter = int(uid_pathname.read_text())
            except ValueError:
                pass  # TODO: raise SystemExit
        else:
            uid_pathname.write_text(str(uid_counter))
    #   self.ec_producer.update("uid_counter", uid_counter)  # TODO: REMOVE

    # Create new link to an existing category or dependency
    #
    # aiko_storage_file link new_entry_name entry_path_existing

    def link(self, entry_path_new, entry_path_existing):
        if not Path(entry_path_existing).exists():
            print(
                f'Error: target "{entry_path_existing}" does not exist',
                file=sys.stderr)
            return 1
        resolved_path = Path(entry_path_existing).resolve()
        cwd = Path.cwd()
        if _STORAGE_FILENAME not in resolved_path.parts:
            print(
                f'Error: "{entry_path_existing}" not in the storage directory',
                file=sys.stderr)
            return 1
        base_directory = Path(entry_path_new).parent or Path(".")
        root = base_directory / _ROOT_FILENAME
        link_base = root if root.is_symlink() else Path(_ROOT_FILENAME)

        # Determine path under ".root/storage/"
        try:
            relative_path = resolved_path.relative_to(cwd / _STORAGE_FILENAME)
            relative_storage_path =  \
                f"{_STORAGE_FILENAME}/{relative_path.as_posix()}"
        except ValueError:
            relative_storage_path = resolved_path.relative_to(cwd).as_posix()

        # Construct full path under .root
        dot_root_path = link_base / relative_storage_path
        relative_path = self._relative_path(
            str(dot_root_path), str(base_directory))
        try:
            Path(entry_path_new).symlink_to(relative_path)
        except FileExistsError:
            print(f'Error: "{entry_path_new}" already exists', file=sys.stderr)
            return 1

    # List categories and dependencies in a tree-style recursive format
    #
    # aiko_storage_file list [-c] [-l] [-r] [ENTRY_PATH]

    # TODO: [-c] Show Category entry count
    # TODO: [-r] Set recursion limit (default: 2)
    # TODO: [-u] Show entry UIDS

    # TODO: Hyperspace: list(self, ..., ENTRY_PATH=None, recursive=False):
    # TODO: Hyperspace:     cat, entry_name = _category_traverse(ENTRY_PATH)
    #
    # TODO: Load Dependency information ...
    #       service_filter=None, lifecycle_manager_url=None, storage_url=None

    def list(self, topic_path_response, entry_path=None,
        long_format=False, recursive=False, entry_records=None):

        if not isinstance(long_format, bool):
            long_format = long_format.lower() in ("true", "t")
        if not isinstance(recursive, bool):
            recursive = recursive.lower() in ("true", "t")

    #   category, entry_name = self._category_traverse(entry_path)
    #   if category and entry_name in category.entries:
    #       entry = category.entries[entry_name]
    #       if entry.service_filter.protocol == CATEGORY_PROTOCOL:
    #           category = entry

    #   def _file_count(link_directory):
    #       target_path = Path(link_directory).resolve()
    #       return sum(1 for file in target_path.rglob("*")  \
    #                     if file.is_file() and file.stat().st_size > 0)

        def _get_entry_records(entry_records, entry_path, level=0):
            if level > 0:
                entry_record = [-level, entry_path]
                if topic_path_response:
                    entry_record = generate(entry_record, [])[2:-2]  # TODO: Fix
                entry_records.append(entry_record)
            _traverse_entries(entry_records, entry_path, level,
                _handler_entry_records)
            _traverse_entries(entry_records, entry_path, level,
                _handler_next_level)

    #   def _get_uid_path(link):
    #       absolute_target_path = Path(link).resolve()
    #       try:
    #           storage_path = Path.cwd() / _STORAGE_FILENAME
    #           relative_path = absolute_target_path.relative_to(storage_path)
    #       except ValueError:
    #           relative_path = ""
    #       return str(relative_path)

        def _handler_entry_records(entry_records, entry, level):
        #   uid_path = _get_uid_path(entry.name)  # TODO: Show "uid_path"
            protocol = CATEGORY_PROTOCOL if entry.is_dir() else "*"

        #   service_filter = ServiceFilter("*", entry.name, protocol)
            service_filter = ["*", entry.name, protocol, "*", "*", []]
            lcm_url = None                 # lifecycle_manager_url
            storage_url = None
        #   dependency = DependencyImpl.create_dependency(...)
            dependency = [service_filter, lcm_url, storage_url]
            entry_record = [level, entry.name, dependency]

            if topic_path_response:
                entry_record = generate(entry_record, [])[2:-2]     # TODO: Fix
            entry_records.append(entry_record)

        def _handler_next_level(entry_records, entry, level):
            if entry.is_dir():
            #   if entry_count:
            #       count = _file_count(entry)
            #       if count > 0:
            #           entry_records.append(f" ({count})")
                if recursive:
                    _get_entry_records(entry_records, entry, level+1)

        def _traverse_entries(entry_records, entry_path, level, entry_handler):
            path = Path(entry_path)
            if path.is_dir():
                for entry in sorted(path.iterdir()):
                    if entry.is_symlink() and not entry.name.startswith("."):
                        entry_handler(entry_records, entry, level)

        entry_path = entry_path if entry_path else "."
        _entry_records = []
    #   _get_entry_records(_entry_records, category, None, 0)  # TODO: Later ?
        _get_entry_records(_entry_records, entry_path)
        if entry_records is None:
            CategoryImpl._list_publish(
                topic_path_response, _entry_records, long_format)
        else:
            entry_records += _entry_records

    # Remove Category (directory) or Dependency (file)
    #
    # aiko_storage_file remove entry  # Category or Dependency

    # TODO: Remove should become like an "unlink", not a "destroy"

    def remove(self, entry_name):
        self.destroy(entry_name)

    def __str__(self):
        result = ""
        for entry_name, entry in self.entries.items():
            entry_type = "/" if entry.is_type("Category") else ""
            new_line = "\n" if result else ""
            result += f"{new_line}  {entry_name}{entry_type}"
        return result

    # Update Category (directory) or Dependency (file) details
    #
    # aiko_storage_file update entry_name ...

    def update(self, entry_name, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

        print("### StorageFileImpl.update() unimplemented ###")

# --------------------------------------------------------------------------- #
# Storage file-system based: Low-level implementation

    @classmethod
    def _check_root_symbolic_link(cls):
        root_pathname = Path.cwd() / _ROOT_FILENAME
        if not os.path.lexists(root_pathname):
            raise FileExistsError(
                f'Storage: symbolic link "./{_ROOT_FILENAME}" must exist')
        else:
            if not os.path.islink(root_pathname):
                raise FileExistsError(
                    f'Storage: File "./{_ROOT_FILENAME}" is not a symbolc link')

    # Clean up empty storage directories up to root

    def _clean_storage(self, tracked_directory):
        tracked_path = Path(tracked_directory)
        storage_path = Path(_STORAGE_FILENAME)
        while tracked_path != storage_path and tracked_path != Path("."):
            if tracked_path.is_dir() and not any(tracked_path.iterdir()):
                tracked_path.rmdir()
                tracked_path = tracked_path.parent
            else:
                break

    # Generate a unique storage path avoiding collisions

    def _create_unique_path(self):
        storage_path = Path(_STORAGE_FILENAME)
        existing_paths = set()
        if Path(_TRACKED_PATHNAME).exists():
            existing_paths.update(
                Path(_TRACKED_PATHNAME).read_text().splitlines())
        while True:
            uid = self._generate_uid()
            path_parts = [uid[i:i+2] for i in range(0, len(uid), 2)]
            new_path = storage_path.joinpath(*path_parts)
            if not new_path.exists() and str(new_path) not in existing_paths:
                return str(new_path)

    # Generate UID, persisting the "uid_counter" when using incremental mode

    def _generate_uid(self):
        if os.environ.get("STORAGE_RANDOM_UID","true").lower() in ("true", "t"):
            uid = hashlib.sha256(os.urandom(16)).hexdigest()
        else:
            uid_counter = self.share["uid_counter"]
            uid = str(uid_counter).zfill(_UID_LENGTH)
            uid_counter += 1
            self.ec_producer.update("uid_counter", uid_counter)
            Path(_UID_PATHNAME).write_text(f"{uid_counter}\n")
        return uid[- _UID_LENGTH:]

    # Normalize path to be relative from CWD

    def _normalize_path(self, target_path):
        absolute_target_path = Path(target_path).resolve()
        cwd = Path.cwd()
        try:
            relative_path = absolute_target_path.relative_to(cwd)
            return f"./{relative_path}"
        except ValueError:
            return str(absolute_target_path)

    # Compute relative path using os.path.relpath

    def _relative_path(self, target_path, start_path):
        return os.path.relpath(target_path, start=start_path)

    # Track a created storage path

    def _track_path(self, path):
        relative_path = self._normalize_path(path)
        Path(_TRACKED_PATHNAME).parent.mkdir(parents=True, exist_ok=True)
        with open(_TRACKED_PATHNAME, "a+") as tracked_paths_file:
            tracked_paths_file.seek(0)
            tracked_paths = tracked_paths_file.read().splitlines()
            if relative_path not in tracked_paths:
                tracked_paths_file.write(relative_path + "\n")

    # Untrack a removed storage path

    def _untrack_path(self, path):
        relative_path = self._normalize_path(path)
        tracked_path = Path(_TRACKED_PATHNAME)
        if tracked_path.is_file():
            tracked_paths = tracked_path.read_text().splitlines()
            tracked_path.write_text(
                "\n".join(tp for tp in tracked_paths if tp != relative_path))

# --------------------------------------------------------------------------- #
# Storage file-system based: CLI for Bootstrap commands

@click.group()

def main():
    """Create, Read/List, Update and Destroy Storage"""

    subcommand = click.get_current_context().invoked_subcommand
    subcommands_which_work_anywhere =  \
        ["create", "exit", "initialize", "run"]
        # TODO: add, destroy, dump, link, list, remove, update ?

    if subcommand not in subcommands_which_work_anywhere:
        diagnostic = None
        try:
            StorageFileImpl._check_root_symbolic_link()
        except FileExistsError as file_exists_error:
            diagnostic = str(file_exists_error)
        if diagnostic:
            advice = 'Consider running "aiko_storage_file initialize"'
            raise SystemExit(f"{diagnostic}\n{advice}")

# aiko_storage_file dump [--sort_by_name]

@main.command(name="dump", help="Dump Storage entries and paths")       
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--sort_by_name", "-s", default=False, is_flag=True,
    help="Sort by name")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def dump_command(bootstrap, sort_by_name, storage_name):
    "Dump StorageFile linked entries and tracked_paths"

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.dump(sort_by_name)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.dump(sort_by_name), terminate=True)
        aiko.process.run()

# aiko_storage_file initialize

@main.command(name="initialize")
@click.argument("storage_url", default=_CWD_URL)

def initialize_command(storage_url):
    """Initialize Storage (.root and storage/)

    aiko_storage_file initialize [STORAGE_URL]

    \b
    • STORAGE_URL: Storage file-system location, e.g file://./
    """

    storage_name = get_hostname()
    storage = StorageFileImpl.create_storage(storage_name, storage_url)
    storage.initialize()

# --------------------------------------------------------------------------- #
# Storage file-system based: CLI for Distributed Actor commands

@main.command(name="add", no_args_is_help=True)
@click.argument("dependency_name", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def add_command(dependency_name, bootstrap, storage_name):
    """Add Dependency

    aiko_storage_file add DEPENDENCY_NAME [-b] [-sn STORAGE_NAME]

    \b
    • DEPENDENCY_NAME: Dependency name
    """

    dependency = None                      # TODO: Implement "dependency"

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.add(dependency_name, dependency)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.add(dependency_name, dependency),
            terminate=True)
        aiko.process.run()

@main.command(name="create", no_args_is_help=True)
@click.argument("category_name", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def create_command(category_name, bootstrap, storage_name):
    """Create Category

    aiko_storage_file create CATEGORY_NAME [-b] [-sn STORAGE_NAME]

    \b
    • CATEGORY_NAME: Category name
    """

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.create(category_name)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.create(category_name), terminate=True)
        aiko.process.run()

@main.command(name="destroy", no_args_is_help=True)
@click.argument("entry_name", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def destroy_command(entry_name, bootstrap, storage_name):
    """Destroy Category or Dependency

    aiko_storage_file destroy ENTRY_NAME [-b] [-sn STORAGE_NAME]

    \b
    • ENTRY_NAME: Entry name (either Category or Dependency)
    """

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.destroy(entry_name)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.destroy(entry_name), terminate=True)
        aiko.process.run()

@main.command(name="exit", help="Exit Storage")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def exit_command(storage_name):
    do_command(Storage, ServiceFilter(name=storage_name, protocol=PROTOCOL),
        lambda storage: storage.exit(), terminate=True)
    aiko.process.run()

@main.command(name="link", no_args_is_help=True)
@click.argument("entry_path_new", type=str, required=True, default=None)
@click.argument("entry_path_existing", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def link_command(entry_path_new, entry_path_existing, bootstrap, storage_name):
    """Create new symbolic link to an existing entry

    aiko_storage_file link NEW_PATH EXISTING_PATH [-b] [-sn STORAGE_NAME]

    \b
    • NEW_PATH:      New Entry path
    • EXISTING_PATH: Entry path that already exists
    """

    dependency = None                      # TODO: Implement "dependency"

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.link(entry_path_new, entry_path_existing)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.link(entry_path_new, entry_path_existing),
            terminate=True)
        aiko.process.run()

@main.command(name="list")
@click.argument("entry_path", type=str, required=False, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")
@click.option("--long_format", "-l", is_flag=True,
    help="Long format with Service, LifeCycleManager URL, Storage URL")
@click.option("--recursive", "-r", is_flag=True,
    help="Recursively list Category entries")

def list_command(entry_path, bootstrap, storage_name, long_format, recursive):
    """List Storage Entries

    aiko_storage_file list ENTRY_PATH [-b] [-sn STORAGE_NAME] [-l] [-r]

    \b
    • ENTRY_PATH: Entry path
    """

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.list(None, entry_path, long_format, recursive, None)
    else:
        CategoryImpl.list_command(
            storage_name, entry_path, long_format, recursive, PROTOCOL)
        aiko.process.run()

@main.command(name="remove", no_args_is_help=True)
@click.argument("entry_name", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def remove_command(entry_name, bootstrap, storage_name):
    """Remove Category or Dependency

    aiko_storage_file remove ENTRY_NAME [-b] [-sn STORAGE_NAME]

    \b
    • ENTRY_NAME: Entry name (either Category or Dependency)
    """

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        storage.remove(entry_name)
    else:
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.remove(entry_name), terminate=True)
        aiko.process.run()

@main.command(name="run")
@click.argument("storage_url", default=_CWD_URL)  # TODO: "storage_url" ???????
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")

def run_command(storage_name, storage_url):
    """Run Storage backed by the file-system

    aiko_storage_file run [-sn STORAGE_NAME] [STORAGE_URL]

    \b
    • STORAGE_URL: Storage file-system location, e.g file://./
    """

    storage_name = storage_name if storage_name else get_hostname()
    storage = StorageFileImpl.create_storage(storage_name, storage_url)
    aiko.process.run()  # TODO: Actor:aiko.process.run() or storage.run() ?

@main.command(name="update", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--bootstrap", "-b", default=False, is_flag=True,
    help="Operate standalone (no Storage Service)")
@click.option("--storage_name", "-sn", type=str, default=None,
    help="Storage name, default is the local hostname")
@click.option("--new_entry_name", "-n", type=str, default="None",
    help="New entry name")  
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def update_command(entry_path, bootstrap, storage_name,
    new_entry_name, lifecycle_manager_url, storage_url):        

    """Update StorageFile Entry

    aiko_storage_file update [-sn STORAGE_NAME] ENTRY_PATH

    \b
    • ENTRY_PATH: Entry path
    """

    service_filter = None                  # TODO: Implement "service_filter"

    if bootstrap:
        storage_url = None                 # TODO: Implement "storage_url" ?
        storage = StorageFileImpl.create_storage("<no_name>", storage_url)
        do_command(Storage,
            ServiceFilter(name=storage_name, protocol=PROTOCOL),
            lambda storage: storage.update(entry_path, None,
                service_filter, lifecycle_manager_url, storage_url),
            terminate=True)
        aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



# Usage
# ~~~~~
# pytest [-s] unit/test_stream_lock.py
# pytest [-s] unit/test_stream_lock.py::test_stream_lock
#
# test_stream_lock()
# - See https://github.com/geekscape/aiko_services/pull/42
#
# To Do
# ~~~~~
# - Improve Aiko Services Process exit
#
# - Test StreamEvent.OKAY, StreamEvent.STOP, StreamEvent.ERROR for ...
#   - create_stream, _create_frame_generator, process_frame, destroy_stream
#   - Consolidate any overlap with "test_stream_event.py" ?

import threading
import time
from typing import Tuple

import aiko_services as aiko
from aiko_services.tests.unit import do_create_pipeline

PIPELINE_DEFINITION = """{
  "version": 0, "name": "p_test", "runtime": "python",
  "graph": ["(StreamLock_0 StreamLock_1 StreamLock_2)"],
  "elements": [
    { "name":   "StreamLock_0", "input":  [], "output": [],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_stream_lock" }
      }
    },
    { "name":   "StreamLock_1", "input":  [], "output": [],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_stream_lock" }
      }
    },
    { "name":   "StreamLock_2", "input":  [], "output": [],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_stream_lock" }
      }
    }
  ]
}
"""
# TODO: Does this need to be a DataSource to replicate the original problem ?

class StreamLock_0(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("stream_lock:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
        self.logger.info(f"{self.my_id()} start_stream()")
        return aiko.StreamEvent.OKAY, {}

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"{self.my_id()} process_frame()")
        return aiko.StreamEvent.OKAY, {}

# NOTE: This causes deliberately causes a StreamEvent.ERROR
#       in either start_stream() or process_frame()

class StreamLock_1(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("stream_lock:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
        self.logger.info(f"{self.my_id()} start_stream()")
        diagnostic = "Deliberate aiko.StreamEvent.ERROR"
        return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
    #   return aiko.StreamEvent.OKAY, {}

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"{self.my_id()} process_frame()")
        diagnostic = "Deliberate aiko.StreamEvent.ERROR"
        return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
    #   return aiko.StreamEvent.OKAY, {}

# NOTE: This is here to show whether a subsequent PipelineElement is processed

class StreamLock_2(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("stream_lock:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
        self.logger.info(f"{self.my_id()} start_stream()")
        return aiko.StreamEvent.OKAY, {}

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"{self.my_id()} process_frame()")
        aiko.process.terminate()  # TODO: Improve Aiko Services Process exit
        return aiko.StreamEvent.OKAY, {}

# NOTE: This has nothing to do with the test-case, here to ensure process exit

def thread_hack():
    time.sleep(0.5)
#   print()
#   print(f"{'Name':<15} {'ID':<10} {'Alive':<7} {'Daemon'}")
#   print("-" * 45)
#   for t in threading.enumerate():
#       print(f"{t.name:<15} {t.ident!s:<10} {t.is_alive():<7} {t.daemon}")
    aiko.process.terminate()

def test_stream_lock():
    threading.Thread(target=thread_hack).start()
    do_create_pipeline(PIPELINE_DEFINITION)



#!/usr/bin/env python3
#
# Aiko Service: HyperSpace(Category(Dependency))
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# HyperSpace is a LifeCycleManager of Categories (category.py and lifecycle.py)
# Create, Read, Update and Destroy a network graph of Categories
# Add to, List and Remove from any Category, which is a group of Entries
# Entries can be either a Category or a Dependency
# Category paths support a tree-like structure "category_a/category_b/entry_c"
#
# Can use either "aiko_hyperspace" (preferred) or "./hyperspace.py"
#
# Usage: Distributed Actor
# ~~~~~~~~~~~~~~~~~~~~~~~~
# aiko_hyperspace.py run     [storage_url]  # file in-memory sqlite mqtt valkey
# aiko_hyperspace.py dump
# aiko_hyperspace.py exit
#
# aiko_hyperspace.py add     ENTRY_PATH
# aiko_hyperspace.py create  CATEGORY_PATH
# aiko_hyperspace.py destroy CATEGORY_PATH
# aiko_hyperspace.py link    ENTRY_PATH_NEW ENTRY_PATH_EXIST    # TODO: "link"
# aiko_hyperspace.py list    [ENTRY_PATH] [--long_format|-l] [--recursive|-r]
# aiko_hyperspace.py remove  ENTRY_PATH_EXIST                   # TODO: "unlink"
# aiko_hyperspace.py update  ENTRY_PATH -p protocol -t key_0=value_0
#
# Usage: Bootstrap
# ~~~~~~~~~~~~~~~~
# export STORAGE_RANDOM_UID=False  # debug using predictable incrementing UIDs
#
# aiko_storage_file initialize [STORAGE_URL]  # set-up a file-system directory
# aiko_storage_file dump -b [--sort_by_name]  # show low-level storage files
# aiko_storage_file list -b [-l] [-r] [PATH]  # show file-system files
#
# To Do
# ~~~~~
# - Separate the design and implementation of HyperSpace and Storage
#   - HyperSpace: create(): Category with different HyperSpace root ?
#   - Storage:    create(): Category or Dependency (Definition and Contents)
#
# * HyperSpace/Registrar(Category): Filters(Code), Order(Sort), Pagination
#
# - Implement "aiko_hyperspace run --register_storage_service" optional flag
# - Implement "aiko_hyperspace repl" for interactive CRUD, etc
#
# * FIX: Consolidate service.py:ServiceFilter into ServiceFields (and use this)
#
# * FIX: "service.py:service_args()" to include "owner" --> "run_command()"
#
# * FIX: AttributeError: 'ServiceRemoteProxy' object has no attribute 'create'
#        do_command(HyperSpace, ServiceFilter(...),  # But HyperSpaceImpl works
#            lambda hyperspace: hyperspace.create(), ...)      # create() fails
#
# * HyperSpace as-a CategoryManager as-a LifeCycleManager as-a Category
#
# * HyperSpace/StorageFile uses FileSystemEventPatternMatch in "scheme_file.py"
#
# * aiko_hyperspace export output_filename.hyperspace --> hyperspace commands
#
# * Validate HyperSpace structure, see StorageFile structure validation
#
# * Registrar(Actor) requires Category plus implement "(update)" for all fields
#
# * HyperSpace(Actor) requires Category and Dependencies
#   - Services that implement Category assign tag "category=true" (temporary)
#
# * Aiko Dashboard plug-in for HyperSpace(Actor), e.g REPL and tree view
#
# * Implement Category "owner" field and populate automatically
#   * Implement "destroy --force" flag to destroy other owner's Categories
#   * System Categories owned by "aiko"
#
# * Consider only one primary HyperSpace per host (refactor Registrar code)
#   * Don't allow two HyperSpace with the same name
# * HyperSpace primary / watchdog: monitor and relaunch primary
# * Unify HyperSpace on different hosts in the same namespace ?
#     * All the home/office servers plus handle mobile laptops (discovery) ?
#
# * HyperSpace uses Leases to enable lazy Category loading, running, unloading
#   - "aiko_hyperspace run" option for Category lease-time and maximum loaded
#   - ProcessManager(LifeCycleManager): Complete Service LifeCycle State
#     - Improvements to "src/aiko_services/main/state.py"
#   - ProcessManager(LifeCycleManager): Lazy loading of HyperSpace entries
#
# - StorageManager (content, definition) backed by ...
#   - ❌ In-memory, ✅ File-system, ❌ SQLite3, ❌ MQTT
# - HyperSpace (links) backed by ...
#   - ❌ In-memory, ✅ File-system, ❌ SQLite3, ❌ MQTT, ❌ Knowledge Graph

from abc import abstractmethod
import click

from aiko_services.main import *
from aiko_services.main.storage import StorageFileImpl
from aiko_services.main.utilities import dir_base_name, get_hostname

__all__ = ["HyperSpace"]

ACTOR_TYPE = "hyperspace"
VERSION = 0
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{VERSION}"

_CWD_URL = "file://./"  # relative URL for the current working directory
_PATH_DELIMITER = "/"
# _RESPONSE_TOPIC = f"{aiko.topic_in}"

# --------------------------------------------------------------------------- #

class HyperSpace(Category, Actor):
    Interface.default(
        "HyperSpace", "aiko_services.main.hyperspace.HyperSpaceImpl")

    @abstractmethod
    def create(self, category_path):
        pass

    @abstractmethod
    def destroy(self, category_path):
        pass

    @abstractmethod
    def dump(self):
        pass

# --------------------------------------------------------------------------- #

class HyperSpaceImpl(HyperSpace):
    @classmethod
    def create_hyperspace(cls, hyperspace_name=None,
        storage_url=_CWD_URL, protocol=PROTOCOL, tags=None, transport=None):

        tags = list(tags) + ["ec=true"]
        init_args = actor_args(
            hyperspace_name, None, None, protocol, tags, transport)
        init_args["storage_url"] = storage_url
        return compose_instance(HyperSpaceImpl, init_args)

    def __init__(self, context, storage_url=_CWD_URL):
        context.call_init(self, "Actor", context)
        context.call_init(self, "Category", context)

        self.category = context.get_implementation("Category")        # methods

        self.share.update({                             # Inherit from Category
            "source_file": f"v{VERSION}⇒ {__file__}",
            "storage_url": storage_url,
            "metrics": {
                "created": 0,
                "running": 1 + len(self.entries)   # Including self 😅
            #   "time_started": 0            # TODO: UTC time started --> Actor
            },                               #       or time.monotonic()
        })
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

        self._hyperspace_load(storage_url)

    def _category_path_iterator(self, category_path, delimiter=_PATH_DELIMITER):
        for name in category_path.split(delimiter):
            yield name

    def _category_traverse(self, category_path, delimiter=_PATH_DELIMITER):
        dirname, basename = dir_base_name(category_path)
        current_category = self
        if dirname == ".":
            return current_category, basename

        for category_name in self._category_path_iterator(dirname):
            if category_name in current_category.entries:
                entry = current_category.entries[category_name]
                if entry.service_filter.protocol != CATEGORY_PROTOCOL:
                    return None, None
                current_category = entry
            else:
                return None, None
        return current_category, basename

    # Default outcome is to use StorageFile to persist the new Dependency.
    # "_hyperspace_load()" only needs to create the in-memory representation
    # from Storage persistence, hence "use_storage=False"

    def add(self, entry_path,
        service_filter, lifecycle_manager_url=None, storage_url=None,
        use_storage=True):

        category, entry_name = self._category_traverse(entry_path)
        if category:
            if service_filter and not isinstance(service_filter, ServiceFilter):
                service_filter = ServiceFilter(*service_filter)
            if service_filter.name == "*":
                service_filter.name = entry_name
            self.category.add(category, entry_name,
                service_filter, lifecycle_manager_url, storage_url)
            dependency = compose_instance(DependencyImpl, dependency_args(
                None, service_filter, lifecycle_manager_url, storage_url))

            if use_storage:
                self.storage.add(entry_name, dependency)

    # Default outcome is to use StorageFile to persist the new Category.
    # "_hyperspace_load()" only needs to create the in-memory representation
    # from Storage persistence, hence "use_storage=False"

    def create(self, category_path, use_storage=True):
        current_category = self
        category_created = False
        tags = ["ec=true"]

        for category_name in self._category_path_iterator(category_path):
            if category_name in current_category.entries:
                entry = current_category.entries[category_name]
                if entry.service_filter.protocol != CATEGORY_PROTOCOL:
                    break
                category = entry
            else:
                init_args = actor_args(
                    category_name, None, None, CATEGORY_PROTOCOL, tags)
                init_args["service_filter"] = ServiceFilter(
                    name=category_name, protocol=CATEGORY_PROTOCOL, tags=[])
                category = compose_instance(CategoryImpl, init_args)
                current_category.entries[category_name] = category
                category_created = True

                if use_storage:
                    self.storage.create(str(category_path))

            current_category = category

        if category_created:
            self.ec_producer.update("entries", len(self.entries))

# TODO: Categories destroy() versus Dependencies remove() ?  Check protocol ?
# TODO: Recursively remove Categories ... using "limit" argument
# TODO: Replace "destroy()" with "remove()", only destroy when last reference
# TODO: Remove Dependencies and clean-up any resources ?

    def destroy(self, category_path):
        category_parent, category_name = self._category_traverse(category_path)
        self._destroy(category_parent, category_name)

    def _destroy(self, category_parent, category_name):
        if category_parent and category_name in category_parent.entries:
            category = category_parent.entries[category_name]
            if category.service_filter.protocol == CATEGORY_PROTOCOL:
                del category_parent.entries[category_name]
                aiko.process.remove_service(category.service_id)
                self.ec_producer.update("entries", len(self.entries))

    def dump(self):
        if len(self.entries):
            state = f" ...\n{self}"
        else:
            state = ": no entries"
        self.logger.info(f"Dump state{state}")

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            self.logger.setLevel(str(item_value).upper())

    def exit(self):
        aiko.process.terminate()

    def _hyperspace_load(self, storage_url=_CWD_URL):
        entry_records = []
        storage_name = f"{get_hostname()}_hs"
        storage_url = None                 # TODO: Implement "storage_url" ?
        self.storage = StorageFileImpl.create_storage(
            storage_name, storage_url, register_service=False)
        self.storage.list(None, None, False, True, entry_records)

        metrics = [0, 0]  # [Dependencies, Categories]
        entry_path = ""
        for record in entry_records:
            level = int(record[0])
            entry_name = record[1]
            if level < 0:
                entry_path = entry_name
            else:
                if record[2][0]:
                    service_filter = ServiceFilter(*record[2][0])
                    is_category = service_filter.protocol == CATEGORY_PROTOCOL
                else:
                    service_filter = ServiceFilter(name=entry_name)
                    is_category = False
                if entry_path:
                    entry_name = f"{entry_path}/{entry_name}"
                if is_category:
                    self.create(entry_name, use_storage=False)
                else:
                    self.add(entry_name,
                        service_filter, None, None, use_storage=False)
                metrics[1 if is_category else 0] += 1
        print(f"Dependencies: {metrics[0]}, Categories: {metrics[1]}")

    def is_type(self, type_name):
        if type_name.lower() == "hyperspace":
            return True
        return self.category.is_type(self, type_name)

    def list(self, topic_path_response, entry_path=None,
        long_format=False, recursive=False, entry_records=None):

        if not isinstance(long_format, bool):
            long_format = long_format.lower() in ("true", "t")
        if not isinstance(recursive, bool):
            recursive = recursive.lower() in ("true", "t")

        category, entry_name = self._category_traverse(entry_path)
        if category and entry_name in category.entries:
            entry = category.entries[entry_name]
            if entry.service_filter.protocol == CATEGORY_PROTOCOL:
                category = entry

        def get_entry_records(entry_records, category, category_name, level):
            if level > 0:
                entry_records += [f"{-level} {category_name}"]
            entry_records += category._get_entry_records(None, level)
            if recursive:
                for entry_name, entry in category.entries.items():
                    if entry.is_type("Category"):
                        get_entry_records(
                            entry_records, entry, entry_name, level+1)

        entry_records = []
        get_entry_records(entry_records, category, None, 0)
        CategoryImpl._list_publish(
            topic_path_response, entry_records, long_format)

# TODO: Categories destroy() versus Dependencies remove() ?  Check protocol ?

    def remove(self, entry_path):
        category_parent, entry_name = self._category_traverse(entry_path)
        if category_parent and entry_name in category_parent.entries:
            entry = category_parent.entries[entry_name]
            if entry.service_filter.protocol == CATEGORY_PROTOCOL:
                self._destroy(category_parent, entry_name)
            else:
                self.category.remove(category_parent, entry_name)

    def __str__(self):
        result = ""
        for entry_name, entry in self.entries.items():
            entry_type = "/" if entry.is_type("Category") else ""
            new_line = "\n" if result else ""
            result += f"{new_line}  {entry_name}{entry_type}"
        return result

    def update(self, entry_path, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

        category_parent, entry_name = self._category_traverse(entry_path)
        if category_parent and entry_name in category_parent.entries:
            self.category.update(category_parent, entry_name, service,
                service_filter, lifecycle_manager_url, storage_url)

# --------------------------------------------------------------------------- #
# HyperSpace CLI: Distributed Actor commands
#
# To Do
# ~~~~~
# add
# ~~~
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# * Created and modified time
# - Owner defaults to current user, ACLs "u:name g:name o:name rwx:name ..." ?
# - ServiceFilter fields (see aiko_hyperspace create)
#
# create
# ~~~~~~
# * Differentiate Categories with same name, e.g ENTRY_PATH tag ?
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# * Created and modified time
# - Owner defaults to current user, ACLs "u:name g:name o:name rwx:name ..." ?
# - ServiceFilter fields ...
#   -n,   --service_name TEXT           # Service name
#   -p,   --protocol TEXT               # CATEGORY_PROTOCOL (default)
#   -tr,  --transport TEXT              # Service transport (MQTT default)
# * -o,   --owner TEXT                  # Service owner     (current user)
#   -t,   --tags TEXT                   # Service tags
# * -lcm, --lifecycle_manager_url TEXT  # LifeCycleManager URL
# * -s,   --storage_url TEXT            # Storage URL (file, sqlite, MQTT, etc)
#
# destroy
# ~~~~~~~
# * Clean-up all Process, Service, Actor, Dependency, Category resources
# * Only destroy() Categories with a LifeCycleManager of "HyperSpaceImpl.self"
# - Regex, e.g wildcards "*", "+", ...
# - [r]ecursive [level]  # default: 1, * means no limit
# - Only owner can destroy ... implement ACLs later
#
# list
# ~~~~
# * Specific "entry_path"
# * Regex, e.g wildcards "*", "+", ...
# * [c]ount  # Category Entry count
# * [l]ong_format
# * [r]ecursive [level]  # default: 1 or * ? (* means no limit)
#
# remove
# ~~~~~~
# - Regex, e.g wildcards "*", "+", ...
# - [r]ecursive [level]  # default: 1, * means no limit
# - Only owner can remove ... ACLs later
#
# update
# ~~~~~~
# * category_path tree: category/category/category
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# - Regex, e.g wildcards "*", "+", ...
# - ServiceFilter fields (see aiko_hyperspace create)
# - Only owner can update ... implement ACLs later

@click.group()

def main():
    """Create, Read/List, Update and Destroy HyperSpace"""

    subcommand = click.get_current_context().invoked_subcommand
    subcommands_which_work_anywhere = [
        "add", "create", "destroy", "dump", "exit", "list", "remove", "update"]

    if subcommand not in subcommands_which_work_anywhere:
        diagnostic = None
        try:
            StorageFileImpl._check_root_symbolic_link()
        except FileExistsError as file_exists_error:
            diagnostic = str(file_exists_error)
        if diagnostic:
            advice = 'Consider running "aiko_storage initialize"'
            raise SystemExit(f"{diagnostic}\n{advice}")

@main.command(name="add", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--service_name", "-n", type=str, default="*",
    help="Service name")
@click.option("--protocol", "-p", type=str, default="*",
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="*",
    help="Service transport")
@click.option("--owner", "-o", type=str, default="*",
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def add_command(entry_path, hyperspace_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Add HyperSpace Entry

    aiko_hyperspace add ENTRY_PATH

    \b
    • ENTRY_PATH: Entry PATH
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(Category,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.add(entry_path,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

@main.command(name="create", no_args_is_help=True)
@click.argument("category_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def create_command(category_path, hyperspace_name):
    """Create HyperSpace Category

    aiko_hyperspace create [-hn HYPERSPACE_NAME] CATEGORY_PATH

    \b
    • CATEGORY_PATH: Category path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.create(category_path), terminate=True)
    aiko.process.run()

@main.command(name="destroy", no_args_is_help=True)
@click.argument("category_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def destroy_command(category_path, hyperspace_name):
    """Destroy HyperSpace Category

    aiko_hyperspace destroy [-hn HYPERSPACE_NAME] CATEGORY_PATH

    \b
    • CATEGORY_PATH: Category path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.destroy(category_path), terminate=True)
    aiko.process.run()

@main.command(name="dump", help="Dump HyperSpace state")
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def dump_command(hyperspace_name):
    do_command(HyperSpace,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.dump(), terminate=True)
    aiko.process.run()

@main.command(name="exit", help="Exit HyperSpace")
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def exit_command(hyperspace_name):
    do_command(HyperSpace,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.exit(), terminate=True)
    aiko.process.run()

@main.command(name="list")
@click.argument("entry_path", type=str, required=False, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--long_format", "-l", is_flag=True,
    help="Long format with Service, LifeCycleManager URL, Storage URL")
@click.option("--recursive", "-r", is_flag=True,
    help="Recursively list Category entries")

def list_command(entry_path, hyperspace_name, long_format, recursive):
    """List HyperSpace Entries

    aiko_category list ENTRY_PATH [-l] [-r]

    \b
    • ENTRY_PATH: Entry path
    """

    CategoryImpl.list_command(
        hyperspace_name, entry_path, long_format, recursive, PROTOCOL)
    aiko.process.run()

@main.command(name="remove", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def remove_command(entry_path, hyperspace_name):
    """Remove HyperSpace Entry

    aiko_hyperspace remove [-hn HYPERSPACE_NAME] ENTRY_PATH

    \b
    • ENTRY_PATH: Entry path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.remove(entry_path), terminate=True)
    aiko.process.run()

@main.command(name="run")
@click.argument("storage_url", required=False, default=None)  # TODO: Implement
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--protocol", "-p", type=str, default=PROTOCOL,
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default=None,
    help="Service transport")
# @click.option("--owner", "-o", type=str, default=None,      # TODO: Implement
#     help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")

def run_command(storage_url, hyperspace_name, protocol, transport, tags):
    """Run HyperSpace

    aiko_hyperspace run [-hn HYPERSPACE_NAME] [STORAGE_URL]

    \b
    • STORAGE_URL: HyperSpace storage location, e.g file://...
    """

    hyperspace_name = hyperspace_name if hyperspace_name else get_hostname()
    hyperspace = HyperSpaceImpl.create_hyperspace(
        hyperspace_name, storage_url, protocol, tags, transport)
    aiko.process.run()

@main.command(name="update", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--service_name", "-n", type=str, default="0:",     # None --> 0:
    help="Service name")
@click.option("--protocol", "-p", type=str, default="0:",         # None --> 0:
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="0:",       # None --> 0:
    help="Service transport")
@click.option("--owner", "-o", type=str, default="0:",            # None --> 0:
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def update_command(entry_path, hyperspace_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Update HyperSpace Entry

    aiko_hyperspace update [-hn HYPERSPACE_NAME] ENTRY_PATH

    \b
    • ENTRY_PATH: Entry path
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.update(entry_path, None,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



"""Highlighter Runtime
===========================

This module provides the core runtime infrastructure for the Highlighter SDK, including
the main Runtime class that orchestrates agent execution, signal handling, configuration
management, and data processing workflows.

The Runtime class serves as the primary entry point for executing Highlighter agents,
handling various input sources (files, URLs, raw data), and managing the lifecycle
of processing tasks.

Key Components:
    - Runtime: Main class for orchestrating agent execution
    - Signal handling for graceful shutdown and configuration reloading
    - Network retry logic with circuit breaker patterns
    - Data source parsing and validation
    - Integration with Aiko Services framework

Example:
    Basic usage of the Runtime class:

    >>> from highlighter.core.runtime import Runtime
    >>> runtime = Runtime(agent_definition="path/to/agent.json")
    >>> runtime.run(files=["path/to/image.jpg"])
"""

from __future__ import annotations

import logging
import os
import signal
import sys
import threading
import time
from contextlib import suppress
from datetime import datetime
from pathlib import Path
from queue import Queue
from typing import Iterable, List, Optional, Union
from urllib.parse import urlparse
from uuid import UUID

import aiko_services as aiko
import aiohttp
import requests
from aiko_services import Stream
from gql.transport.exceptions import TransportError
from pybreaker import CircuitBreaker, CircuitBreakerError
from tenacity import (
    after_log,
    before_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

# FIXME: (JOSH) the 'core' namespace should not require any other highlighter
# namespaces. If a module that is in 'core' needs to access namespaces outside
# of 'core' it should not be in 'core'.
import highlighter.core.decorators as decorators
from highlighter.agent.agent import HLAgent
from highlighter.cli.cli_logging import log_queue_size
from highlighter.client.gql_client import HLClient, close_all_thread_local_clients
from highlighter.client.tasks import TaskContext, lease_task
from highlighter.core.config import (
    HighlighterRuntimeConfig,
)
from highlighter.core.shutdown import runtime_stop_event
from highlighter.core.thread_watch import dump_stacks, join_all, patch_threading

# ────────────────────────────────────────────
# Section 1: helpers originally inside _start
# ────────────────────────────────────────────

# Runtime configuration constants
THREAD_GRACEFUL_TIMEOUT = 5  # seconds per thread on shutdown


def _log_to_stderr_fallback(message: str) -> None:
    """Best-effort write used when structured logging might not reach stdout."""

    stream = getattr(sys, "stderr", None)
    if stream is None:
        return
    with suppress(Exception):
        stream.write(f"core.runtime: {message}\n")
        stream.flush()


def _make_network_fn_decorator(config, logger):
    """Create a decorator for network calls with retry logic and per-thread circuit breaker.

    This function creates a decorator that adds robust error handling to network
    operations, including exponential backoff retries and per-thread circuit breaker
    patterns to prevent cascading failures while allowing concurrent operations.

    Args:
        config: HighlighterRuntimeConfig instance with network settings
        logger: Logger instance for retry/failure reporting

    Returns:
        Decorator function that can be applied to network-calling functions

    The decorator handles:
        - Exponential backoff retry with configurable max attempts
        - Per-thread circuit breaker pattern to isolate failures
        - Logging of retry attempts and failures
        - Multiple exception types (Transport, Connection, Timeout errors)

    Note:
        Each thread gets its own circuit breaker instance. This prevents a failure
        in one thread from blocking network calls in other threads, which is
        critical for concurrent upload operations.
    """
    # Thread-local storage for per-thread circuit breakers
    _thread_local = threading.local()

    def get_thread_breaker():
        """Get or create circuit breaker for current thread."""
        if not hasattr(_thread_local, "circuit_breaker"):
            _thread_local.circuit_breaker = CircuitBreaker(
                fail_max=1,  # consider if we need adjustable fail max config.network.fail_max. Set to 1 to fail, and retry after reset_timeout
                reset_timeout=config.network.reset_timeout,
            )
        return _thread_local.circuit_breaker

    def decorator(fn):
        retry_decorated = retry(
            wait=wait_exponential(multiplier=0.2, max=10),
            stop=stop_after_attempt(config.network.max_retries),
            retry=retry_if_exception_type(
                (
                    TransportError,
                    aiohttp.ClientConnectionError,  # aiohttp network errors
                    requests.ConnectionError,  # requests network errors
                    TimeoutError,  # Python built-in TimeoutError
                )
            ),
            before=before_log(logger, logging.DEBUG),
            after=after_log(logger, logging.DEBUG),
        )(fn)

        def wrapped(*args, **kwargs):
            last_exception = None

            def capture_and_call(*args, **kwargs):
                nonlocal last_exception
                try:
                    return retry_decorated(*args, **kwargs)
                except Exception as e:
                    last_exception = e
                    raise

            # Get circuit breaker for this specific thread
            breaker = get_thread_breaker()

            try:
                return breaker.call(capture_and_call, *args, **kwargs)
            except CircuitBreakerError as cbe:
                # If we have a captured exception, re-raise with it as the second argument
                if last_exception is not None:
                    raise CircuitBreakerError(
                        f"Circuit breaker opened due to: {last_exception}", last_exception
                    ) from last_exception
                # Otherwise, propagate the original circuit breaker error
                raise

        return wrapped

    return decorator


class Runtime:
    """Core runtime class for executing Highlighter agents.

    The Runtime class orchestrates the execution of Highlighter agents, managing
    configuration, signal handling, and agent lifecycle.
    It supports multiple execution modes including stream initiation and task polling.

    Attributes:
        logger: Logger instance for runtime operations
        agent_definition: Path to agent definition file
        dump_definition: Optional path for dumping agent definition
        allow_non_machine_user: Whether to allow non-machine user tasks
        hl_cfg: Runtime configuration instance
        hl_client: Highlighter client for API communication
        queue_response: External response queue for test harness/CLI integration.
            This is distinct from HLAgent.process_task's per-task response queue.
    """

    def __init__(
        self,
        agent_definition: Union[str, dict, os.PathLike],
        dump_definition: Optional[str] = None,
        allow_non_machine_user: bool = False,
        hl_cfg: Optional[HighlighterRuntimeConfig] = None,
        hl_client=None,
        # External response queue (test harness / CLI). Distinct from
        # HLAgent.process_task's per-task response queue.
        queue_response=None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize the Runtime instance.

        Args:
            agent_definition: Path to the agent definition JSON file
            dump_definition: Optional path to dump processed agent definition
            allow_non_machine_user: Allow processing tasks from non-machine users
            hl_cfg: Pre-configured runtime configuration
            hl_client: Highlighter client for API communication
            queue_response: External response queue for test/CLI stream integration.
                This is not the per-task response queue used by HLAgent.process_task.
            logger: Optional logger instance
        """

        self._install_signal_handlers()
        patch_threading()

        self.agent = None
        self.agent_definition = agent_definition
        self.dump_definition = dump_definition
        self.allow_non_machine_user = allow_non_machine_user
        self.hl_client = hl_client
        self.queue_response = queue_response

        self.load_config()
        self.hl_cfg = hl_cfg if hl_cfg is not None else self.hl_cfg  # override with passed in config

        root_logger = logging.getLogger()
        if not root_logger.handlers:
            raise RuntimeError(
                "No logging handlers are configured. Call configure_root_logger(...) before creating Runtime."
            )

        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

    # ─────────────────────── signal handling section ───────────────────────
    def _complete_queued_work_then_shutdown(self, signum, frame):
        """Handle SIGTERM signal for graceful shutdown.

        This handler prevents new streams from being created and allows
        current queued work to complete before stopping the agent.

        Args:
            signum: Signal number received
            frame: Current stack frame (unused)
        """
        name = signal.Signals(signum).name
        msg = f"{name} received – preventing new streams, and stopping agent after current queued work"
        self._log_signal(level=logging.INFO, message=msg)
        self.agent.disable_create_stream()
        # Stop all streams after processing process_frame calls on the Aiko event queue
        self.agent.stop_all_streams(graceful=True)

    def _interrupt_and_shutdown(self, signum, frame):
        """Handle SIGINT signal for immediate shutdown.

        This handler stops all streams immediately and drains worker threads.
        Currently processing frames will complete but no new work will start.

        Args:
            signum: Signal number received
            frame: Current stack frame (unused)
        """
        # Restore the default SIGINT handler in case our handler hangs
        signal.signal(signum, signal.SIG_DFL)
        name = signal.Signals(signum).name
        msg = f"{name} received – stopping all streams, draining worker threads and stopping agent"
        self._log_signal(level=logging.INFO, message=msg)
        self.agent.disable_create_stream()
        # Stop all streams immediately.
        # NOTE: If a frame is currently being processed it won't be interrupted.
        self.agent.stop_all_streams(graceful=False)

    def _quick_abort(self, signum, frame):
        """Handle SIGABRT/SIGQUIT signals for emergency abort.

        This handler dumps thread stacks for debugging and then allows
        the default signal behavior (usually core dump) to proceed.

        Args:
            signum: Signal number received (SIGABRT or SIGQUIT)
            frame: Current stack frame (unused)
        """
        name = signal.Signals(signum).name
        msg = f"{name} received – dumping stacks, exiting HARD!"
        self._log_signal(level=logging.ERROR, message=msg)
        dump_stacks(level=logging.ERROR)
        # Re-raise default behaviour so the OS still gets a core dump / stack trace
        signal.signal(signum, signal.SIG_DFL)
        signal.raise_signal(signum)

    def reload_config(self, signum=None, frame=None):
        """Reload configuration at runtime.

        Can be called explicitly or as a signal handler (SIGHUP).
        Reloads the HighlighterRuntimeConfig from disk.

        Args:
            signum: Signal number (when used as signal handler)
            frame: Current stack frame (when used as signal handler)
        """
        self.logger.info("Reloading configuration...")
        try:
            self.load_config(force_reload=True)
            self.logger.info("Reloaded config complete.")
        except Exception as e:
            self.logger.error(f"Error reloading config: {e}")

    def _install_signal_handlers(self) -> None:
        """Register OS-level signal handlers.

        Registers handlers for graceful shutdown, configuration reload,
        and emergency abort. Must be called from the main thread.

        Signals handled:

        - SIGINT: Immediate shutdown
        - SIGTERM: Graceful shutdown after current work
        - SIGQUIT/SIGABRT: Emergency abort with diagnostics
        - SIGHUP: Configuration reload (Unix only)
        """
        signal.signal(signal.SIGINT, self._interrupt_and_shutdown)
        signal.signal(signal.SIGTERM, self._complete_queued_work_then_shutdown)
        signal.signal(signal.SIGQUIT, self._quick_abort)
        signal.signal(signal.SIGABRT, self._quick_abort)
        try:  # not present on Windows
            signal.signal(signal.SIGHUP, self.reload_config)
        except AttributeError:
            pass

    def load_config(self, *, force_reload: bool = False):
        """Load or reload the runtime configuration.

        Loads HighlighterRuntimeConfig from the standard configuration
        sources. Updates the instance's hl_cfg attribute.

        Raises:
            Exception: If configuration loading fails
        """
        self.hl_cfg = HighlighterRuntimeConfig.load(force_reload=force_reload)

    def start(self):
        """Start the runtime and initialize the agent.

        This method:

        1. Clears any previous stop events
        2. Loads configuration
        3. Sets up network decorators
        4. Creates and starts the HLAgent in a new thread
        """
        self.start_time = datetime.now()
        self.logger.info(f"Starting runtime on pid {os.getpid()}")
        runtime_stop_event.clear()

        decorators.network_fn_decorator = _make_network_fn_decorator(self.hl_cfg, self.logger)

        self.agent = HLAgent(
            self.agent_definition,
            dump_definition=self.dump_definition,
            timeout_secs=self.hl_cfg.agent.timeout_secs,
            task_lease_duration_secs=self.hl_cfg.agent.task_lease_duration_secs,
            task_polling_period_secs=self.hl_cfg.agent.task_polling_period_secs,
        )
        self.logger.info("Starting agent thread …")
        self.agent.run_in_new_thread()

    def shutdown(self):
        """Shutdown the runtime gracefully.

        This method:

        1. Stops all agent streams
        2. Stops the agent
        3. Sets the runtime stop event
        4. Dumps thread stacks for debugging
        5. Waits for all threads to join with timeout
        """
        if self.agent is not None:
            self.agent.stop_all_streams()
            self.agent.stop()
        runtime_stop_event.set()  # broadcast the shutdown request

        # Cleanup resources
        close_all_thread_local_clients()

        dump_stacks(level=logging.INFO)
        join_all(timeout=THREAD_GRACEFUL_TIMEOUT)

        elapsed = datetime.now() - self.start_time
        # Using timedelta's string representation is more readable than fractional days.
        msg = f"Runtime shutdown complete. Ran for {elapsed}"
        self._log_signal(level=logging.INFO, message=msg)

    def run(
        self,
        server: bool = False,
        stream_definitions: None | List[dict] = None,
        files: None | List[str | Path] = None,
        urls: None | List[str] = None,
        step_id: None | UUID = None,
        task_ids: None | List[UUID] = None,
    ) -> None:
        """Execute the runtime's main processing loop.

        This is the primary entry point for runtime execution. It:

        1. Starts the agent
        2. Initiate process inputs if specified
        3. Initiate processing tasks if specified
        4. Shuts down gracefully after completion

        Execution Modes:

        - Stream initiation: When files, urls, or stream_definitions are provided
        - Task polling: When step_id is provided
        - Specific tasks: When task_ids is provided
        """

        if step_id and task_ids:
            raise ValueError("--step-id and --task-ids are mutually exclusive")

        self.start()  # start agent

        if stream_definitions is None:
            stream_definitions = []
        if urls is None:
            urls = []
        if files is not None:
            for filename in files:
                file_url = f"file://{filename}"
                urls.append(file_url)
        for url in urls:
            stream_definitions.append({"data_sources": f"({len(url)}:{url})"})

        stream_ids = []
        task = TaskContext()
        for stream_id, stream_parameters in enumerate(stream_definitions or []):
            stream_id = str(stream_id)
            stream_ids.append(stream_id)
            # Extract subgraph_name from stream_parameters if present
            stream_params = stream_parameters.copy()
            graph_path = stream_params.pop("subgraph_name", None)
            # Pass the external queue_response (if provided) so callers can observe
            # stream events and frame outputs for non-task streams.
            self.agent.create_stream(
                stream_id=stream_id,
                graph_path=graph_path,
                parameters=stream_params,
                queue_response=self.queue_response,
            )
            self.agent.pipeline.stream_leases[stream_id].stream.variables["task"] = task

        # Wait for streams to stop
        iterations = 0

        if step_id:
            self.agent.poll_for_tasks_loop(step_id, allow_non_machine_user=self.allow_non_machine_user)

        elif task_ids:
            if not self.allow_non_machine_user:
                self.agent.check_user_is_machine()
            # TODO: load hl client if not passed in
            for task_id in [t.strip() for t in task_ids]:
                task = lease_task(
                    self.hl_client,
                    task_id=task_id,
                    lease_sec=self.hl_cfg.agent.task_lease_duration_secs,
                    set_status_to="RUNNING",
                )
                self.agent.process_task(task)

        if server:
            self.logger.info("Continuing to run in server mode")
            while True:
                time.sleep(1)
        else:
            while True:
                # process queue response for frame_id
                # process events - drain all events for frame_id
                if not any([stream_id in self.agent.pipeline.stream_leases for stream_id in stream_ids]):
                    break

                task.finalise_submissions_on_recording_state_off()

                iterations += 1
                if iterations % 10 == 0:
                    self.logger.info("Waiting for streams to stop")
                    if self.queue_response is not None:
                        log_queue_size(
                            self.logger,
                            logging.INFO,
                            self.queue_response,
                            "Stream output queue (external) size: %(qsize)s",
                            error_message="Unable to read external queue size",
                        )

                time.sleep(1)

        self.logger.info("Finished processing, joining worker threads")
        self.shutdown()

    def _log_signal(self, *, level: int, message: str) -> None:
        if self.logger.isEnabledFor(level):
            self.logger.log(level, message)
        else:
            _log_to_stderr_fallback(message)

    def create_stream(
        self,
        stream_id,
        graph_path=None,
        parameters=None,
        grace_time=None,
        queue_response=None,
        topic_response=None,
        frame_complete_hook=None,
        destroy_stream_hook=None,
    ):
        if self.agent is None:
            raise ValueError("Start runtime via runtime.start() before creating a stream")
        return self.agent.create_stream(
            stream_id=stream_id,
            graph_path=graph_path,
            parameters=parameters,
            grace_time=grace_time,
            queue_response=queue_response,
            topic_response=topic_response,
            frame_complete_hook=frame_complete_hook,
            destroy_stream_hook=destroy_stream_hook,
        )

    def destroy_stream(
        self, stream_id, graceful=False, use_thread_local=True, acquire_lock=True, diagnostic={}
    ):
        if self.agent is None:
            raise ValueError("Cannot destroy streams when agent is not running")
        return self.agent.destroy_stream(
            stream_id=stream_id,
            graceful=graceful,
            use_thread_local=use_thread_local,
            acquire_lock=acquire_lock,
            diagnostic=diagnostic,
        )

    def create_frame(
        self, stream: Stream, frame_data: dict, frame_id: int | None = None, graph_path: str | None = None
    ):
        if self.agent is None:
            raise ValueError("Cannot create frame when agent is not running")
        return self.agent.create_frame(stream, frame_data, frame_id=frame_id, graph_path=graph_path)



import aiko_services as aiko
from aiko_services.main.pipeline import (
    _PIPELINE_HOOK_DESTROY_STREAM,
)
from aiko_services.main.stream import StreamEvent, StreamState
from aiko_services.tests.unit import do_create_pipeline

PIPELINE_DEFINITION = """{
  "version": 0, "name": "p_test", "runtime": "python",
  "graph": ["(A)"],
  "elements": [
    { "name":   "A",
      "input":  [{ "name": "a", "type": "int" }],
      "output": [{ "name": "a", "type": "int" }],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_pipeline_exception_handling" }
      }
    }
  ]
}
"""

class A(aiko.PipelineElement):
    def __init__(self, context):
        context.call_init(self, "PipelineElement", context)

    def frame_generator(self, stream, frame_id):
        if frame_id == 0:
            return StreamEvent.OKAY, {"a": 1}
        else:
            return StreamEvent.STOP, {}

    def process_frame(self):
        return StreamEvent.OKAY, {}

    def start_stream(self, stream, stream_id):
        raise Exception("TEST")
        return StreamEvent.OKAY, None

    def stop_stream(self, stream, stream_id):
        self.stop()
        return StreamEvent.OKAY, None


def test_pipeline_hooks():
    def hook_destroy_stream(hook_name, component, logger, variables, options):
        assert variables["stream"].state == StreamState.ERROR

    hooks = {
        _PIPELINE_HOOK_DESTROY_STREAM: hook_destroy_stream,
    }
    do_create_pipeline(PIPELINE_DEFINITION, hooks=hooks)



from typing import Any, Dict, List, Tuple
from uuid import UUID

from highlighter import DatumSource, Entity, Observation
from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.core.data_models import DataSample


class MockClassifier(Capability):
    """Mock classifier that creates a dummy observation"""

    class StreamParameters(Capability.StreamParameters):
        # When we specify the output taxonomy of the element in the
        # pipeline element outputs then we can get it from there. For now
        # we add it to the parameters
        output_attribute_id: UUID
        output_enum_id: UUID

    def process_frame(
        self, stream, entities: Dict[UUID, Entity], data_samples: List[DataSample]
    ) -> Tuple[StreamEvent, dict]:
        parameters = self.stream_parameters(stream.stream_id)
        for entity in entities.values():
            for annotation in entity.annotations:
                datum_source = DatumSource(confidence=1, frame_id=annotation.datum_source.frame_id)
                Observation(
                    annotation=annotation,
                    attribute_id=parameters.output_attribute_id,
                    value=parameters.output_enum_id,
                    datum_source=datum_source,
                )
        return StreamEvent.OKAY, {"entities": entities}



import hashlib
import os
import re
from contextlib import contextmanager


def sha512_of_content(content):
    sha512 = hashlib.sha512()
    sha512.update(content)
    return sha512.hexdigest()


def get_slug(name) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


def stringify_if_not_null(attribute):
    return str(attribute) if attribute is not None else None


@contextmanager
def change_dir(path):
    """Context manager that changes to a directory and returns to the original on exit.

    Args:
        path: Path to change to (str or Path)

    Example:
        with change_dir('/tmp'):
            # Do work in /tmp
            print(os.getcwd())  # /tmp
        # Back to original directory
    """
    original_dir = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(original_dir)



# To Do
# ~~~~~
# - Replace "Gst.parse_launch()" with hand-built pipeline !

import sys
from threading import Thread

if sys.version_info >= (3,0):
  from queue import Queue, Empty
else:
  from Queue import Queue, Empty

from aiko_services.elements.gstreamer import *

__all__ = ["VideoFileWriter"]

# -----------------------------------------------------------------------------

class VideoFileWriter:
  def __init__(self, filename, width, height, framerate):
    self.Gst = gst_initialise()
    self.queue = Queue(maxsize=30)

# vtenc_h264: 1280x768, 512x288, 256x144
    gst_launch_command = "appsrc name=source ! videoconvert ! videoscale ! videorate ! video/x-raw,width=1280,height=768,framerate=25/1 ! {} ! splitmuxsink location={}"
#   location={}_%03d.mp4
#   max-files=10
#   max-size-bytes=1024000
#   max-size-time=1000000  # 1 second

#   gst_launch_command = "appsrc name=source ! {} ! mp4mux ! filesink location={}"
#   gst_launch_command = "appsrc name=source ! jpegenc ! avimux ! filesink location={}"
#   gst_launch_command = "appsrc name=source ! {} ! decodebin ! mpegtsmux ! filesink location={}"

    pipeline = self.Gst.parse_launch(gst_launch_command.format(utilities.get_h264_encoder(), filename))

    appsrc = pipeline.get_child_by_name("source")
    src_caps = "video/x-raw,format={},width={},height={},framerate={}".format(utilities.get_format(), width, height, framerate)
    appsrc.props.caps = self.Gst.caps_from_string(src_caps)

    appsrc.props.stream_type  = 0  # GST_APP_STREAM_TYPE_STREAM
    appsrc.props.is_live      = True
    appsrc.props.do_timestamp = True
    appsrc.props.format       = self.Gst.Format.TIME

    pipeline.set_state(self.Gst.State.PLAYING)

    def _run():
      while True:
        frame = self.queue.get()
        if "type" in frame and frame["type"] == "EOS":
          appsrc.emit('end-of-stream')
        else:
#         print("VFW frame: ", frame["id"], self.queue.qsize())
          data = frame["image"].tobytes()
          buffer = self.Gst.Buffer.new_allocate(None, len(data), None)
          buffer.fill(0, data)
          appsrc.emit('push-buffer', buffer)

    self._t = Thread(target=_run, args=())
    self._t.daemon = True
    self._t.start()

  def queue_size(self):
    return self.queue.qsize()

  def write_frame(self, frame):
    self.queue.put(frame)



from importlib.metadata import version as get_installed_version

__all__ = ["__version__"]

try:
    __version__ = get_installed_version("highlighter-sdk")
except Exception as e:
    __version__ = "unable to determine version, see the pyproject.toml"



"""NeuFlowV2 optical flow estimation with built-in preprocessing."""

import logging
import os
from pathlib import Path
from typing import Optional

import numpy as np
import onnxruntime
import requests
import tqdm

__all__ = ["NeuFlowV2", "flow_to_image"]

MAX_KERNEL_SIZE = 7

AVAILABLE_MODELS = ["neuflow_mixed", "neuflow_sintel", "neuflow_things"]
MODEL_BASE_URL = "https://github.com/ibaiGorordo/ONNX-NeuFlowV2-Optical-Flow/releases/download/0.1.0"


def _download_model(url: str, path: str):
    """Download a model from URL."""
    print(f"Downloading model from {url} to {path}")
    r = requests.get(url, stream=True, timeout=30)
    with open(path, "wb") as f:
        total_length = int(r.headers.get("content-length", 0))
        for chunk in tqdm.tqdm(
            r.iter_content(chunk_size=1024 * 1024),
            total=max(1, total_length // (1024 * 1024)),
            bar_format="{l_bar}{bar:10}",
        ):
            if chunk:
                f.write(chunk)
                f.flush()


def _get_preprocessed_path(original_path: str) -> str:
    """Get path for preprocessed model variant."""
    path = Path(original_path)
    return str(path.parent / f"{path.stem}_preprocessed{path.suffix}")


def _ensure_preprocessed_model(original_path: str) -> str:
    """Ensure preprocessed model exists, creating it if necessary."""
    from ._neuflow_preprocessing import prepend_preprocessing_to_neuflow

    original_path = str(original_path)
    preprocessed_path = _get_preprocessed_path(original_path)

    # Download original model if needed
    if not os.path.exists(original_path):
        Path(original_path).parent.mkdir(parents=True, exist_ok=True)
        model_name = Path(original_path).stem
        if model_name not in AVAILABLE_MODELS:
            raise ValueError(f"Unknown model: {model_name}. Available: {AVAILABLE_MODELS}")
        url = f"{MODEL_BASE_URL}/{model_name}.onnx"
        _download_model(url, original_path)

    # Create preprocessed version if needed
    if not os.path.exists(preprocessed_path):
        print(f"Creating preprocessed model: {preprocessed_path}")
        prepend_preprocessing_to_neuflow(original_path, preprocessed_path)

    return preprocessed_path


class NeuFlowV2:
    """
    Optical flow estimator with built-in preprocessing.

    The model includes Gaussian blur and resize operations, with the blur
    kernel configurable at runtime. On first use, automatically downloads
    the base model and creates a preprocessed version.

    Args:
        path: Path to the NeuFlow ONNX model (preprocessed version auto-created)
        use_gpu: Whether to use GPU for inference
        blur_kernel_size: Gaussian blur kernel size (3, 5, 7) or None for no blur
    """

    def __init__(self, path: str, use_gpu: bool = True, blur_kernel_size: Optional[int] = 3):
        self.logger = logging.getLogger(__name__)

        # Ensure preprocessed model exists (downloads and creates if needed)
        preprocessed_path = _ensure_preprocessed_model(path)

        # Select execution provider
        if use_gpu:
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
        else:
            providers = ["CPUExecutionProvider"]

        self.session = onnxruntime.InferenceSession(preprocessed_path, providers=providers)
        self.logger.info(f"NeuFlowV2 initialized with providers: {providers}")

        # Set initial blur kernel
        self.set_blur_kernel(blur_kernel_size)

    def set_blur_kernel(self, kernel_size: Optional[int]):
        """
        Set Gaussian blur kernel size.

        Args:
            kernel_size: Kernel size (3, 5, 7) or None to disable blur
        """
        if kernel_size is None:
            self._blur_kernel = self._make_identity_kernel()
            self.logger.info("Blur disabled (identity kernel)")
        else:
            if kernel_size > MAX_KERNEL_SIZE:
                raise ValueError(f"Max kernel size is {MAX_KERNEL_SIZE}, got {kernel_size}")
            if kernel_size % 2 == 0:
                raise ValueError(f"Kernel size must be odd, got {kernel_size}")
            sigma = (kernel_size - 1) / 6.0
            self._blur_kernel = self._make_gaussian_kernel(kernel_size, sigma)
            self.logger.info(f"Blur kernel set to {kernel_size}x{kernel_size} (sigma={sigma:.2f})")

    @staticmethod
    def _make_gaussian_kernel(size: int, sigma: float) -> np.ndarray:
        """Create Gaussian kernel for depthwise convolution, zero-padded to MAX_KERNEL_SIZE."""
        coords = np.arange(size) - size // 2
        g = np.exp(-(coords**2) / (2 * sigma**2))
        kernel_2d = np.outer(g, g)
        kernel_2d = kernel_2d / kernel_2d.sum()
        # Shape for depthwise conv: (out_channels=3, in_channels/groups=1, H, W)
        kernel_4d = np.tile(kernel_2d[np.newaxis, np.newaxis, :, :], (3, 1, 1, 1))
        # Zero-pad to max kernel size
        if size < MAX_KERNEL_SIZE:
            pad = (MAX_KERNEL_SIZE - size) // 2
            kernel_4d = np.pad(kernel_4d, ((0, 0), (0, 0), (pad, pad), (pad, pad)))
        return kernel_4d.astype(np.float32)

    @staticmethod
    def _make_identity_kernel() -> np.ndarray:
        """Create identity kernel (no-op convolution)."""
        kernel = np.zeros((MAX_KERNEL_SIZE, MAX_KERNEL_SIZE), dtype=np.float32)
        kernel[MAX_KERNEL_SIZE // 2, MAX_KERNEL_SIZE // 2] = 1.0
        return np.tile(kernel[np.newaxis, np.newaxis, :, :], (3, 1, 1, 1))

    def __call__(self, prev_frame: np.ndarray, curr_frame: np.ndarray) -> np.ndarray:
        """
        Compute optical flow between two frames.

        Args:
            prev_frame: Previous frame (H, W, 3) uint8 array
            curr_frame: Current frame (H, W, 3) uint8 array

        Returns:
            Optical flow (H, W, 2) float32 array
        """
        outputs = self.session.run(
            None,
            {
                "prev_image": prev_frame,
                "curr_image": curr_frame,
                "blur_kernel": self._blur_kernel,
            },
        )
        return outputs[0]


# --- Flow visualization utilities ---
# Ref: https://github.com/liruoteng/OpticalFlowToolkit/blob/5cf87b947a0032f58c922bbc22c0afb30b90c418/lib/flowlib.py#L249


def make_color_wheel():
    """
    Generate color wheel according Middlebury color code
    :return: Color wheel
    """
    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR

    colorwheel = np.zeros([ncols, 3])

    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.transpose(np.floor(255 * np.arange(0, RY) / RY))
    col += RY

    # YG
    colorwheel[col : col + YG, 0] = 255 - np.transpose(np.floor(255 * np.arange(0, YG) / YG))
    colorwheel[col : col + YG, 1] = 255
    col += YG

    # GC
    colorwheel[col : col + GC, 1] = 255
    colorwheel[col : col + GC, 2] = np.transpose(np.floor(255 * np.arange(0, GC) / GC))
    col += GC

    # CB
    colorwheel[col : col + CB, 1] = 255 - np.transpose(np.floor(255 * np.arange(0, CB) / CB))
    colorwheel[col : col + CB, 2] = 255
    col += CB

    # BM
    colorwheel[col : col + BM, 2] = 255
    colorwheel[col : col + BM, 0] = np.transpose(np.floor(255 * np.arange(0, BM) / BM))
    col += +BM

    # MR
    colorwheel[col : col + MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
    colorwheel[col : col + MR, 0] = 255

    return colorwheel


colorwheel = make_color_wheel()


def compute_color(u, v):
    """
    compute optical flow color map
    :param u: optical flow horizontal map
    :param v: optical flow vertical map
    :return: optical flow in color code
    """
    [h, w] = u.shape
    img = np.zeros([h, w, 3])
    nanIdx = np.isnan(u) | np.isnan(v)
    u[nanIdx] = 0
    v[nanIdx] = 0

    ncols = np.size(colorwheel, 0)

    rad = np.sqrt(u**2 + v**2)

    a = np.arctan2(-v, -u) / np.pi

    fk = (a + 1) / 2 * (ncols - 1) + 1

    k0 = np.floor(fk).astype(int)

    k1 = k0 + 1
    k1[k1 == ncols + 1] = 1
    f = fk - k0

    for i in range(0, np.size(colorwheel, 1)):
        tmp = colorwheel[:, i]
        col0 = tmp[k0 - 1] / 255
        col1 = tmp[k1 - 1] / 255
        col = (1 - f) * col0 + f * col1

        idx = rad <= 1
        col[idx] = 1 - rad[idx] * (1 - col[idx])
        notidx = np.logical_not(idx)

        col[notidx] *= 0.75
        img[:, :, i] = np.uint8(np.floor(255 * col * (1 - nanIdx)))

    return img


def flow_to_image(flow, maxrad=None):
    """
    Convert flow into middlebury color code image
    :param flow: optical flow map
    :return: optical flow image in middlebury color
    """
    u = flow[:, :, 0]
    v = flow[:, :, 1]

    rad = np.sqrt(u**2 + v**2)
    if maxrad is None:
        maxrad = max(-1, np.max(rad))

    eps = np.finfo(float).eps
    u = np.clip(u, -maxrad + 5, maxrad - 5)
    v = np.clip(v, -maxrad + 5, maxrad - 5)

    u = u / (maxrad + eps)
    v = v / (maxrad + eps)

    img = compute_color(u, v)

    return np.uint8(img)



import logging
import sys
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID

import yaml
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator

from ..core import GQLBaseModel
from .gql_client import HLClient

__all__ = [
    "EvaluationMetric",
    "EvaluationMetricCodeEnum",
    "EvaluationMetricResult",
    "create_evaluation_metric",
    "create_evaluation_metric_result",
    "find_or_create_evaluation_metric",
]


class EvaluationMetricCodeEnum(str, Enum):
    Dice = "Dice"
    mAP = "mAP"
    MaAD = "MaAD"
    MeAD = "MeAD"
    Other = "Other"
    AP = "AP"


class EvaluationMetric(GQLBaseModel):
    model_config = ConfigDict(use_enum_values=True)

    research_plan_id: int
    code: EvaluationMetricCodeEnum
    chart: Optional[str] = None
    description: Optional[str] = None
    iou: Optional[float] = None
    name: str
    object_class_uuid: Optional[Union[UUID, str]] = None
    weighted: Optional[bool] = False
    id: Optional[int] = None

    def dict(self, *args, **kwargs):
        d = super().model_dump(*args, **kwargs)
        if "object_class_uuid" in d:
            d["object_class_uuid"] = str(d["object_class_uuid"])
        return d

    def create(self, client: HLClient):
        class CreateResearchPlanMetricReturnType(GQLBaseModel):
            errors: Any
            research_plan_metric: Optional[EvaluationMetric] = None

        kwargs = EvaluationMetric(
            research_plan_id=self.research_plan_id,
            code=self.code,
            # chart=self.chart,  # ToDo: Add this to the Graphql
            name=self.name,
            description=self.description,
            iou=self.iou,
            weighted=self.weighted,
            object_class_uuid=str(self.object_class_uuid),
        ).gql_dict()

        result = client.createResearchPlanMetric(return_type=CreateResearchPlanMetricReturnType, **kwargs)

        if result.errors:
            raise SystemExit(str(result.errors))
        return result.research_plan_metric

    def result(self, value, training_run_id):
        """Create an evaluation metric result instance without immediately persisting it.

        This method creates an EvaluationMetricResult object that can be used to store
        evaluation results. The result must be explicitly created using the .create()
        method to persist it to Highlighter.

        Args:
            value: The numeric evaluation result value
            training_run_id: ID of the training run that produced this result

        Returns:
            EvaluationMetricResult: A result object ready for creation/persistence

        Note:
            This method replaces the previous create_result() method to provide better
            separation between result object creation and persistence.
        """
        evaluation_metric_result = EvaluationMetricResult(
            research_plan_metric_id=self.id,
            result=value,
            object_class_uuid=self.object_class_uuid,
            training_run_id=training_run_id,
        )
        return evaluation_metric_result


class EvaluationMetricResult(GQLBaseModel, extra="forbid"):
    research_plan_metric_id: int
    result: float
    # iso datetime str will be generated at instantiation
    # if not supplied manually.
    occured_at: Optional[Union[datetime, str]] = Field(
        ..., default_factory=lambda _: datetime.now(timezone.utc).isoformat()
    )
    object_class_uuid: Optional[Union[UUID, str]] = None
    training_run_id: Optional[int] = None

    @field_validator("occured_at", mode="before")
    def set_timestamp(cls, v):
        if v is None:
            v = datetime.now(timezone.utc).isoformat()
        elif isinstance(v, str):
            v = datetime.fromisoformat(v).isoformat()
        elif isinstance(v, datetime):
            v = v.isoformat()
        else:
            raise ValidationError()
        return v

    @classmethod
    def from_yaml(cls, path: Union[Path, str]):
        path = Path(path)
        with path.open("r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    def create(self, client: HLClient):
        class CreateExperimentResultReturnType(BaseModel):
            errors: Any
            experimentResult: Optional[EvaluationMetricResult] = None

        mutation_result: CreateExperimentResultReturnType = client.createExperimentResult(
            return_type=CreateExperimentResultReturnType,
            **self.gql_dict(),
        )
        if mutation_result.errors:
            raise SystemExit(str(mutation_result.errors))

        return mutation_result.experimentResult


def get_existing_evaluation_metrics(client: HLClient, evaluation_id: int):
    class QueryReturnType(GQLBaseModel):
        research_plan_metrics: List[EvaluationMetric]

    query_return_type: QueryReturnType = client.researchPlan(return_type=QueryReturnType, id=evaluation_id)

    return query_return_type.research_plan_metrics


def create_evaluation_metric(
    client: HLClient,
    evaluation_id: int,
    code: Union[EvaluationMetricCodeEnum, str],
    name: str,
    chart: Optional[str] = None,
    description: Optional[str] = None,
    iou: Optional[float] = None,
    weighted: Optional[bool] = False,
    object_class_uuid: Optional[Union[UUID, str]] = None,
) -> EvaluationMetric:

    result = EvaluationMetric(
        research_plan_id=evaluation_id,
        code=code,
        chart=chart,
        name=name,
        description=description,
        iou=iou,
        weighted=weighted,
        object_class_uuid=object_class_uuid,
    ).create(client)

    return result


def find_or_create_evaluation_metric(
    client: HLClient,
    evaluation_metrics: Union[EvaluationMetric, List[EvaluationMetric]],
) -> List[EvaluationMetric]:
    if isinstance(evaluation_metrics, EvaluationMetric):
        _evaluation_metrics = [evaluation_metrics]
    elif isinstance(evaluation_metrics, list):
        _evaluation_metrics = evaluation_metrics
    else:
        raise ValueError(f"Invalid evaluation_metrics, got: {evaluation_metrics}")

    logger = logging.getLogger(__name__)
    evaluation_ids = {e.research_plan_id for e in _evaluation_metrics}
    if len(evaluation_ids) > 1:
        logger.error("All evaluation metrics must have the same research_plan_id")
        sys.exit(1)

    existing_evaluation_metrics = {
        r.name: r for r in get_existing_evaluation_metrics(client, evaluation_ids.pop())
    }
    results: List[EvaluationMetric] = []

    for metric in _evaluation_metrics:
        if metric.name in existing_evaluation_metrics:
            found_metric = existing_evaluation_metrics[metric.name]
            logger.debug(f"Found evaluation metric {found_metric.id}:({found_metric.name})")
            results.append(found_metric)
        else:
            created_metric = metric.create(client)
            logger.debug(f"Created evaluation metric {created_metric.id}:({created_metric.name})")
            results.append(created_metric)

    return results


def create_evaluation_metric_result(
    client: HLClient,
    evaluation_metric_id: int,
    result: Union[float, int],
    occured_at: Optional[Union[datetime, str]] = None,
    object_class_uuid: Optional[Union[str, UUID]] = None,
    training_run_id: Optional[int] = None,
) -> EvaluationMetricResult:
    """Create an evaluation_metric if it does not exist, optionally
    create an evaluation_metric_result if result is not None
    """
    return EvaluationMetricResult(
        research_plan_metric_id=evaluation_metric_id,
        result=result,
        occured_at=occured_at,
        object_class_uuid=object_class_uuid,
        training_run_id=training_run_id,
    ).create(client)



import json
import logging
from pathlib import Path
from typing import List, Tuple, Union
from uuid import uuid4

from PIL import Image

from highlighter.core.geometry import (
    multipolygon_from_coords,
    polygon_from_left_top_width_height_coords,
)

from ....client import PixelLocationAttributeValue
from ....core import OBJECT_CLASS_ATTRIBUTE_UUID
from ...base_models import AttributeRecord, ImageRecord
from ...common import SUPPORTED_IMAGE_EXTENSIONS
from ...interfaces import IReader
from .common import CocoKeys

logger = logging.getLogger(__name__)

PathLike = Union[str, Path]


class CocoReader(IReader):
    """
    Parses a COCO dataset, returning a comprehensive list of its images and attributes.

    This reader treats the local filesystem as the source of truth and handles:
    - Verification of image existence and integrity, automatically skipping missing or corrupt files.
    - Extraction of true image dimensions from disk to supervise potentially inaccurate JSON metadata.
    - Graceful recovery from a malformed geometry or an invalid category identifier.

    Requirements:
    - A valid COCO JSON annotation file.
    - A directory containing the corresponding images.
    """

    format_name = "coco"

    def __init__(
        self,
        json_path: PathLike,
        image_dir: PathLike,
        bbox_only: bool = False,
        fix_invalid_polygons: bool = False,
    ):
        self.json_path = Path(json_path)
        self.image_dir = Path(image_dir)
        self.bbox_only = bbox_only
        self.fix_invalid_polygons = fix_invalid_polygons

        if not self.json_path.is_file():
            raise FileNotFoundError(f"Annotation file not found: {self.json_path}")
        if not self.image_dir.is_dir():
            raise FileNotFoundError(f"Image directory not found: {self.image_dir}")

        self.class_map = {}

    def _extract_spatial_data(self, anno):
        """
        A generator function for ingesting exactly one of at most two spatial
        attributes contained in any single annotation object from the input
        JSON.

        By default, we preference ingestion of the segmentation label over the
        bounding box ("bbox") label. You can optionally override this behavior by
        setting the bbox_only option to True.
        """
        if CocoKeys.SEG in anno and not self.bbox_only:
            try:
                pts = [list(zip(f[:-1:2], f[1::2])) for f in anno[CocoKeys.SEG]]
                poly = multipolygon_from_coords(pts, fix_invalid_polygons=self.fix_invalid_polygons)
                yield PixelLocationAttributeValue.from_geom(poly)
            except Exception as e:
                logger.warning(f"Skipping invalid segmentation in annotation {anno.get('id')}: {e}")

        elif CocoKeys.BBOX in anno:
            try:
                yield PixelLocationAttributeValue.from_geom(
                    polygon_from_left_top_width_height_coords(anno[CocoKeys.BBOX])
                )
            except Exception as e:
                logger.warning(f"Skipping invalid bbox in annotation {anno.get('id')}: {e}")

    def read(self) -> Tuple[List[AttributeRecord], List[ImageRecord]]:
        """
        Parses annotations and images, verifying integrity against the filesystem.

        Returns:
            Tuple[List[AttributeRecord], List[ImageRecord]]: verified image and attribute records.
        """
        try:
            with self.json_path.open() as f:
                data = json.load(f)
        except (OSError, json.JSONDecodeError) as e:
            raise ValueError(f"Failed to load coco annotations from {self.json_path}: {e}")

        # extract class-from-ID map and group annotations by image ID for O(1) lookup
        self.class_map = {c[CocoKeys.ID]: c[CocoKeys.NAME] for c in data.get(CocoKeys.CATS, [])}
        annos_map = {}
        for a in data.get(CocoKeys.ANNOS, []):
            # if this annotation's segmentation field is a polygon list, map to
            # it map to it from the ID of its containing image (for uploading)
            if not a.get("iscrowd"):
                annos_map.setdefault(a[CocoKeys.IMAGE_ID], []).append(a)

        imgs, attrs, seen = [], [], set()

        # iterate over images defined in JSON
        for i in data.get(CocoKeys.IMAGES, []):
            if (f_id := i.get(CocoKeys.ID)) in seen:
                continue

            seen.add(f_id)
            # verify file existence and integrity
            f_path = self.image_dir / i.get(CocoKeys.FILE_NAME, "")

            if f_path.suffix.lower() not in SUPPORTED_IMAGE_EXTENSIONS:
                logger.warning(f"Skipping image with unsupported extension: {f_path}")
                continue

            try:
                if not f_path.exists():
                    logger.warning(f"Image file not found: {f_path}")
                    continue
                with Image.open(f_path) as img:
                    # pixel dimensions
                    width, height = img.size
            except Exception as e:
                logger.error(f"Skipping corrupt or unreadable image {f_path}: {e}")
                continue

            imgs.append(
                ImageRecord(
                    data_file_id=str(f_id),
                    width=width,
                    height=height,
                    filename=str(f_path),
                    split="data",
                    extra_fields=i.get(CocoKeys.EXTRA_FIELDS, {}),
                )
            )

            # process annotations associated with this image
            for a in annos_map.get(f_id, []):
                if (cid := a.get(CocoKeys.CAT_ID)) not in self.class_map:
                    logger.warning(f"Skipping annotation with unknown category id {cid} in image {f_id}")
                    continue

                e_id = a.get(CocoKeys.EXTRA_FIELDS, {}).pop("entity_id", str(uuid4()))
                attrs.append(
                    AttributeRecord(
                        data_file_id=str(f_id),
                        entity_id=e_id,
                        attribute_id=str(OBJECT_CLASS_ATTRIBUTE_UUID),
                        attribute_name=OBJECT_CLASS_ATTRIBUTE_UUID.label,
                        value=self.class_map[cid],
                    )
                )

                # capture all valid geometries (seg AND bbox if both present)
                for g in self._extract_spatial_data(a):
                    attrs.append(AttributeRecord.from_attribute_value(str(f_id), g, entity_id=e_id))

                # add an additional attribute for each extra, 'metadata' field
                for k, v in a.get(CocoKeys.EXTRA_FIELDS, {}).items():
                    attrs.append(
                        AttributeRecord(
                            data_file_id=str(f_id),
                            entity_id=e_id,
                            attribute_id=k,
                            attribute_name="ToDo",
                            value=v,
                        )
                    )

        return imgs, attrs



import logging
import threading
import time
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, List, Optional, Union
from uuid import UUID

from highlighter.cli.cli_logging import COLOURS, RESET, format_timing

from ..core import GQLBaseModel, paginate
from .aws_s3 import upload_file_to_s3
from .base_models import File as FileType
from .base_models import PageInfo
from .gql_client import HLClient

__all__ = [
    "get_data_files",
    "create_data_file",
]

BYTES_PER_MEGABYTE = 1_000_000
BITS_PER_MEGABIT = 1_000_000


def get_data_files(
    client,
    data_file_ids: Optional[List[int]] = None,
    data_file_uuids: Optional[List[Union[str, UUID]]] = None,
    data_source_id: Optional[List[int]] = None,
    data_source_uuid: Optional[List[str]] = None,
    file_hash: Optional[List[str]] = None,
):
    class FileTypeConnection(GQLBaseModel):
        page_info: PageInfo
        nodes: List[FileType]

    kwargs = {
        "id": data_file_ids,
        "uuid": data_file_uuids,
        "dataSourceId": data_source_id,
        "dataSourceUuid": data_source_uuid,
        "fileHash": file_hash,
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    return paginate(client.fileConnection, FileTypeConnection, **kwargs)


def create_data_file(
    client: HLClient,
    data_file_path: Union[str, Path],
    data_source_uuid: UUID,
    site_id: Optional[str] = None,
    observed_timezone: Optional[str] = None,
    recorded_at: Optional[str] = None,
    recorded_until: Optional[datetime] = None,
    metadata: str = "{}",
    uuid: Optional[str] = None,
    multipart_filesize: Optional[str] = None,
    content_type: str = "image",
) -> FileType:
    logger = logging.getLogger(__name__)
    current_thread = threading.current_thread()
    thread_id = current_thread.ident
    thread_name = current_thread.name

    data_file_path = Path(data_file_path)
    if not data_file_path.exists():
        raise FileNotFoundError(f"{data_file_path}")

    # Get file size for context
    file_size_bytes = data_file_path.stat().st_size
    file_size_mb = file_size_bytes / BYTES_PER_MEGABYTE

    # Time the S3 upload
    logger.debug(
        f"[Thread {thread_id}/{thread_name}] Starting S3 upload for {data_file_path.name} ({file_size_mb:.2f} MB)"
    )
    s3_upload_start = time.perf_counter()
    file_data = upload_file_to_s3(
        client,
        str(data_file_path),
        multipart_filesize=multipart_filesize,
        data_source_uuid=str(data_source_uuid),
    )
    s3_upload_elapsed = time.perf_counter() - s3_upload_start
    upload_mbps = (
        (file_size_bytes * 8) / (s3_upload_elapsed * BITS_PER_MEGABIT) if s3_upload_elapsed > 0 else 0
    )
    logger.debug(
        f"[Thread {thread_id}/{thread_name}] S3 upload for {data_file_path.name} took {s3_upload_elapsed:.3f}s"
    )

    if recorded_at is None:
        recorded_at = datetime.now(timezone.utc).isoformat()

    class CreateFileResponse(GQLBaseModel):
        file: Optional[FileType] = None
        errors: Any = None

    # Time the GraphQL API call
    logger.debug(f"[Thread {thread_id}/{thread_name}] Starting GraphQL create_file for {data_file_path.name}")
    api_call_start = time.perf_counter()
    create_data_file_response = client.create_file(
        return_type=CreateFileResponse,
        dataSourceUuid=str(data_source_uuid),
        originalSourceUrl=str(data_file_path),
        fileData=file_data,
        siteId=site_id,
        observedTimezone=observed_timezone,
        recordedAt=recorded_at,
        recordedUntil=recorded_until,
        metadata=metadata,
        uuid=uuid,
        contentType=content_type,
    )
    api_call_elapsed = time.perf_counter() - api_call_start

    # Format thread info with color
    thread_info = f"{COLOURS['grey']}[Thread {thread_id}/{thread_name}]{RESET}"

    # Format timings with colors
    timing_s3 = format_timing("S3", s3_upload_elapsed, "s", color_threshold={"fast": 1.0, "slow": 5.0})
    timing_graphql = format_timing(
        "GraphQL", api_call_elapsed, "s", color_threshold={"fast": 0.5, "slow": 2.0}
    )

    # Single consolidated log line with all timing information
    logger.info(
        f"{thread_info} {data_file_path.name} | " f"{timing_s3}({upload_mbps:.1f}Mbps) {timing_graphql}"
    )

    errors = create_data_file_response.errors
    if errors:
        # If it's a string, just raise it
        if isinstance(errors, str):
            raise ValueError(errors)
        # If it's an iterable (e.g. list of strings)
        elif isinstance(errors, (list, tuple)):
            raise ValueError(". ".join(str(e) for e in errors))
        # If it's something else, convert to string
        else:
            raise ValueError(str(errors))
    return create_data_file_response.file



from typing import Any, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel

from .gql_client import HLClient
from .object_classes import read_object_classes

__all__ = ["create_workflow"]


def create_workflow(
    client: HLClient,
    name: str,
    description: Optional[str] = None,
    parent_id: Optional[int] = None,
    object_class_ids: Optional[List[int]] = None,
    object_class_uuids: Optional[List[Union[str, UUID]]] = None,
):
    class Workflow(BaseModel):
        id: int
        name: str

    class ReturnType(BaseModel):
        errors: Any = None
        workflow: Optional[Workflow] = None

    if (object_class_ids is not None) and (object_class_uuids is not None):
        raise ValueError(f"Cannot have both object_class_ids and object_class_uuids")

    if object_class_uuids is not None:
        uuid_strs = [str(u) for u in object_class_uuids]
        object_classes = read_object_classes(client, uuid=uuid_strs)
        object_class_ids = [o.id for o in object_classes]

    result = client.createWorkflow(
        return_type=ReturnType,
        name=name,
        description=description,
        parentId=parent_id,
        objectClassIds=object_class_ids,
    )

    if result.errors:
        raise ValueError(f"{result.errors}")
    return result.workflow



import time
from queue import Empty, Queue

from highlighter.agent.capabilities import DataSourceCapability, StreamEvent


class QueueDataSource(DataSourceCapability):
    class StreamParameters(DataSourceCapability.StreamParameters):
        queue: Queue

    def process_frame(self, stream, **kwargs):
        return StreamEvent.OKAY, kwargs

    def frame_generator(self, stream, frame_id):
        queue = self.stream_parameters(stream.stream_id).queue
        try:
            frame_data = queue.get(timeout=0.001)  # Somehow this is faster than queue.get(False)?
            if frame_data == "STOP":
                return StreamEvent.STOP, {}
            return StreamEvent.OKAY, frame_data
        except Empty:
            # We need to keep producing empty frames so that the frame generator
            # gives up the stream lock so frame processing can continue
            return StreamEvent.OKAY, {}

    def start_stream(self, stream, stream_id):
        super().start_stream(stream, stream_id)
        self.create_frames(stream, frame_generator=self.frame_generator)
        return StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        queue = self.stream_parameters(stream.stream_id).queue
        queue.put("STOP")
        return super().stop_stream(stream, stream_id)



"""mDNS/Zeroconf discovery helpers shared by the CLI and SDK."""

from __future__ import annotations

import ipaddress
import logging
import platform
import re
import subprocess  # nosec B404 - Required for ARP/ping commands with validated inputs
import threading
import time
from concurrent.futures import ThreadPoolExecutor, wait
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Optional

from zeroconf import ServiceBrowser, ServiceInfo, Zeroconf

LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT_SECONDS = 5  # balances discovery completeness with CLI responsiveness
DEFAULT_SERVICE_TYPES = ["_http._tcp.local.", "_psia._tcp.local.", "_CGI._tcp.local."]
DEFAULT_DEVICE_KEYWORDS = ["HIKVISION", "CAMERA", "AXIS", "DAHUA", "IP CAM", "IPCAM"]


class DiscoveryError(Exception):
    """Raised when the Zeroconf discovery process fails."""


@dataclass
class DiscoveredDevice:
    """Representation of a discovered device."""

    service_name: str
    hostname: str
    ip: str
    port: str
    serial: str
    mac: Optional[str] = None


class DeviceServiceListener:
    """Listener for mDNS service discovery events."""

    def __init__(self, zeroconf: Zeroconf) -> None:
        self.zeroconf = zeroconf
        self.services: Dict[str, ServiceInfo] = {}
        self.lock = threading.Lock()

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        try:
            info = zc.get_service_info(type_, name)
            if info:
                with self.lock:
                    self.services[name] = info
        except Exception as exc:  # pragma: no cover - defensive
            LOGGER.debug("Failed to add service %s of type %s: %s", name, type_, exc)

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        self.add_service(zc, type_, name)

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        with self.lock:
            removed = self.services.pop(name, None)
            if removed is None:
                LOGGER.debug("Attempted to remove unknown service %s", name)

    def get_services(self) -> List[ServiceInfo]:
        with self.lock:
            return list(self.services.values())


def normalize_mac(mac: str) -> Optional[str]:
    """Normalize MAC address format."""

    hex_digits = re.sub(r"[^0-9A-Fa-f]", "", mac)
    if len(hex_digits) != 12:
        return None

    pairs = [hex_digits[i : i + 2].upper() for i in range(0, 12, 2)]
    return ":".join(pairs)


class MacResolver:
    """Resolve MAC addresses for IPs across Linux, macOS, and Windows."""

    def __init__(self, platform_name: Optional[str] = None) -> None:
        self.platform_name = platform_name or platform.system()

    def resolve(self, ip: str) -> Optional[str]:
        if not _is_valid_ip(ip):
            LOGGER.debug("Skipping MAC resolution for invalid IP %s", ip)
            return None
        mac = self._query_arp_tables(ip)
        if mac:
            return mac

        self._ping_ip(ip)
        return self._query_arp_tables(ip)

    def _query_arp_tables(self, ip: str) -> Optional[str]:
        for command in self._commands_for_platform(ip):
            mac = self._run_and_extract(command)
            if mac:
                return mac
        return None

    def _run_and_extract(self, command: List[str]) -> Optional[str]:
        try:
            result = subprocess.run(
                command, capture_output=True, text=True, timeout=5
            )  # nosec B603 - Command uses validated IP address with hardcoded system commands
        except (subprocess.TimeoutExpired, FileNotFoundError) as exc:
            LOGGER.debug("Command %s failed during MAC resolution: %s", command, exc)
            return None

        if result.returncode != 0:
            LOGGER.debug("Command %s exited with %s", command, result.returncode)
            return None

        mac = self._extract_mac(result.stdout)
        if mac:
            return normalize_mac(mac)
        return None

    def _commands_for_platform(self, ip: str) -> Iterable[List[str]]:
        system = self.platform_name
        if system == "Linux":
            yield ["ip", "neighbor", "show", ip]
            yield ["arp", "-n", ip]
        elif system == "Darwin":
            yield ["arp", "-n", ip]
            yield ["arp", ip]
        elif system == "Windows":
            yield ["arp", "-a", ip]
        else:
            yield ["arp", "-n", ip]

    def _ping_ip(self, ip: str) -> None:
        command: List[str]
        system = self.platform_name
        if system == "Linux":
            command = ["ping", "-c", "1", "-W", "1", ip]
        elif system == "Darwin":
            command = ["ping", "-c", "1", ip]
        elif system == "Windows":
            command = ["ping", "-n", "1", "-w", "1000", ip]
        else:
            command = ["ping", "-c", "1", ip]

        try:
            subprocess.run(
                command, capture_output=True, timeout=3
            )  # nosec B603 - Command uses validated IP address with hardcoded ping command
        except (subprocess.TimeoutExpired, FileNotFoundError):
            return

    @staticmethod
    def _extract_mac(output: str) -> Optional[str]:
        pattern = re.compile(r"([0-9A-Fa-f]{2}(?:[:-][0-9A-Fa-f]{2}){5})")
        match = pattern.search(output)
        if match:
            return match.group(1)

        compact = re.compile(r"([0-9A-Fa-f]{12})")
        match = compact.search(output)
        if match:
            return match.group(1)
        return None


def extract_serial_from_name(name: str) -> Optional[str]:
    """Best-effort serial extractor tuned for vendor naming schemes."""

    candidates = re.findall(r"[A-Z0-9-]{8,}", name.upper())
    best: Optional[str] = None
    best_len = 0
    for candidate in candidates:
        candidate = candidate.strip("-")
        if len(candidate) < 8 or len(candidate) > 64:
            continue
        letters = sum(ch.isalpha() for ch in candidate)
        digits = sum(ch.isdigit() for ch in candidate)
        if letters >= 2 and digits >= 2 and len(candidate) > best_len:
            best = candidate
            best_len = len(candidate)
    return best


@dataclass
class DiscoveryConfig:
    """Configuration knobs for device discovery."""

    timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS
    service_types: List[str] = field(default_factory=lambda: list(DEFAULT_SERVICE_TYPES))
    device_keywords: List[str] = field(default_factory=lambda: list(DEFAULT_DEVICE_KEYWORDS))
    max_mac_workers: int = 8


def discover_devices(
    *,
    config: Optional[DiscoveryConfig] = None,
    resolve_macs: bool = True,
    mac_resolver: Optional[MacResolver] = None,
) -> List[DiscoveredDevice]:
    """Discover devices on the network using mDNS/Zeroconf."""

    cfg = config or DiscoveryConfig()

    zeroconf: Optional[Zeroconf] = None
    try:
        zeroconf = Zeroconf()
        listener = DeviceServiceListener(zeroconf)
        for service_type in cfg.service_types:
            ServiceBrowser(zeroconf, service_type, listener)

        time.sleep(cfg.timeout_seconds)

        raw_services = listener.get_services()
        filtered = _filter_device_services(raw_services, cfg.device_keywords)
        devices = _build_devices(filtered)
        deduped = _deduplicate_by_ip(devices)

        if resolve_macs:
            resolver = mac_resolver or MacResolver()
            _resolve_macs(deduped, resolver, cfg)

        return deduped
    except Exception as exc:  # pragma: no cover - exceptional path
        raise DiscoveryError(f"Error during mDNS discovery: {exc}") from exc
    finally:
        if zeroconf is not None:
            zeroconf.close()


def _filter_device_services(services: List[ServiceInfo], keywords: List[str]) -> List[ServiceInfo]:
    keyword_set = {kw.upper() for kw in keywords}
    filtered: List[ServiceInfo] = []
    for info in services:
        service_name = info.name.upper()
        if any(keyword in service_name for keyword in keyword_set):
            filtered.append(info)
    return filtered


def _build_devices(services: List[ServiceInfo]) -> List[DiscoveredDevice]:
    devices: List[DiscoveredDevice] = []
    for info in services:
        addresses = info.parsed_addresses()
        if not addresses:
            continue

        ipv4_addresses = [addr for addr in addresses if ":" not in addr]
        ip = ipv4_addresses[0] if ipv4_addresses else addresses[0]

        if not _is_valid_ip(ip):
            LOGGER.debug("Skipping invalid IP from service %s: %s", info.name, ip)
            continue

        hostname = (info.server or "").rstrip(".")
        port = str(info.port or "80")
        serial = _extract_serial(info.name, hostname)

        devices.append(
            DiscoveredDevice(
                service_name=info.name,
                hostname=hostname,
                ip=ip,
                port=port,
                serial=serial or "",
            )
        )
    return devices


def _deduplicate_by_ip(devices: List[DiscoveredDevice]) -> List[DiscoveredDevice]:
    deduped: Dict[str, DiscoveredDevice] = {}
    for device in devices:
        existing = deduped.get(device.ip)
        if existing:
            if device.serial and not existing.serial:
                existing.serial = device.serial
        else:
            deduped[device.ip] = device
    return list(deduped.values())


def _resolve_macs(devices: List[DiscoveredDevice], resolver: MacResolver, cfg: DiscoveryConfig) -> None:
    with ThreadPoolExecutor(max_workers=cfg.max_mac_workers) as executor:
        futures = [executor.submit(_resolve_single_mac, resolver, device) for device in devices]
        wait(futures)


def _resolve_single_mac(resolver: MacResolver, device: DiscoveredDevice) -> None:
    try:
        device.mac = resolver.resolve(device.ip)
    except Exception as exc:  # pragma: no cover - defensive
        LOGGER.debug("Failed to resolve MAC for %s: %s", device.ip, exc)
        device.mac = None


def _extract_serial(service_name: str, hostname: str) -> Optional[str]:
    for candidate in (hostname, service_name):
        serial = extract_serial_from_name(candidate)
        if serial:
            return serial
    return None


def _is_valid_ip(ip: str) -> bool:
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False



import logging
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from uuid import UUID

import numpy as np

from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.base_models import Entity
from highlighter.client.gql_client import HLClient
from highlighter.client.training_runs import TrainingRunArtefactType
from highlighter.core.data_models import DataSample
from highlighter.datasets.cropping import CropArgs
from highlighter.predictors.onnx_yolov8 import OnnxYoloV8
from highlighter.predictors.output_category_descriptor import OutputCategoryDescriptor

__all__ = ["YoloBoxClassifier"]

logger = logging.getLogger(__name__)


class YoloBoxClassifier(Capability):
    """
    A box classifier capability that uses YOLO to classify existing bounding boxes.

    This capability takes entities with bounding box or polygon annotations from upstream
    (e.g., from an object detector) and classifies each box using a YOLO classifier model.
    The classification results are added as observations to the existing annotations.
    """

    class InitParameters(Capability.InitParameters):
        onnx_file: Optional[str] = None
        training_run_artefact_id: Optional[UUID] = None
        num_classes: int
        class_lookup: Optional[Dict[int | str, OutputCategoryDescriptor]] = None
        conf_thresh: float = 0.1
        nms_iou_thresh: float = 0.5
        is_absolute: bool = True
        cropper: Optional[CropArgs] = None

    def __init__(self, context):
        super().__init__(context)

        onnx_file = self.init_parameters.onnx_file
        if onnx_file is None:
            training_run_artefact_id = self.init_parameters.training_run_artefact_id
            if training_run_artefact_id is None:
                raise ValueError("Must provide `onnx_file` or `training_run_artefact_id`")
            training_run_artefact = HLClient.get_client().trainingRunArtefact(
                return_type=TrainingRunArtefactType, id=training_run_artefact_id
            )
            onnx_file = training_run_artefact.file_url

        artefact_cache_dir = Path.home() / ".cache" / "highlighter" / "agents" / "artefacts"
        self._predictor = OnnxYoloV8(
            onnx_file,
            self.init_parameters.num_classes,
            class_lookup=self.init_parameters.class_lookup,
            conf_thresh=self.init_parameters.conf_thresh,
            nms_iou_thresh=self.init_parameters.nms_iou_thresh,
            is_absolute=self.init_parameters.is_absolute,
            artefact_cache_dir=artefact_cache_dir,
        )

    def process_frame(
        self, stream, data_samples: List[DataSample], entities: Dict[UUID, Entity]
    ) -> Tuple[StreamEvent, Dict]:
        start = time.perf_counter()

        annotations = []
        crop_data_samples = []
        for entity in entities.values():
            for annotation in entity.annotations:
                annotations.append(annotation)
                crop = annotation.crop_image(crop_args=self.init_parameters.cropper)
                crop_data_sample = annotation.data_sample.model_copy()
                crop_data_sample.content = crop
                crop_data_samples.append(crop_data_sample)

        # YOLO classifiers return whole-image annotations
        classification_annotations = self._predictor.predict(crop_data_samples)
        for orig_annotation, class_annotations in zip(annotations, classification_annotations):
            for class_annotation in class_annotations:
                for observation in class_annotation.observations:
                    # Associate the observation with the original annotation
                    observation.annotation = None
                    observation.annotation = orig_annotation

        end = time.perf_counter()
        self.logger.debug(f"YoloBoxClassifier processed in {end - start:.6f} seconds")

        return StreamEvent.OKAY, {"entities": entities}



import logging
from collections import defaultdict
from typing import List, Optional, Union

from highlighter.agent.utilities.entity_writer import (
    EmbeddingAggregationMode,
    IEntityWriter,
)
from highlighter.client.base_models import Entity

logger = logging.getLogger(__name__)

__all__ = ["EntityAggregator"]

PROTO_FRAME_SHELF_LIFE = 50


class EntityAggregator:
    """Aggregates an Entity's annotation.location and annotation.observations
    into tracks.

    An entity can have multiple tracks, see EntityTracks

    Args:
        minimum_track_frame_length (int): Only include tracks in the final
        result if the track has at least this number of frames.

        minimum_embedding_in_track_frame_length (int): Only include aggregate
        embeddings and include them in the final result if the track the at
        least this number of frames.

        embedding_aggregation_mode (EmbeddingAggregationMode): The method used to
        aggregate the embeddings for each frame in a track.
    """

    # TODO do this filtering in the tracker and make configurable
    DEFAULT_MINIMUM_TRACK_FRAME_LENGTH = 1
    DEFAULT_MINIMUM_EMBEDDING_TRACK_FRAME_LENGTH = 4

    def __init__(
        self,
        minimum_track_frame_length=DEFAULT_MINIMUM_TRACK_FRAME_LENGTH,
        minimum_embedding_in_track_frame_length=DEFAULT_MINIMUM_EMBEDDING_TRACK_FRAME_LENGTH,
        embedding_aggregation_mode: EmbeddingAggregationMode = EmbeddingAggregationMode.MOST_REPRESENTATIVE_WITHOUT_OUTLIERS,
        purge_interval: int = 50,
        writer: Optional[IEntityWriter] = None,
    ):

        self._minimum_track_frame_length = minimum_track_frame_length
        self._minimum_embedding_in_track_frame_length = minimum_embedding_in_track_frame_length
        self._proto_tracks = defaultdict(list)
        self._track_entities = defaultdict(list)
        self._purge_interval = purge_interval
        self._next_purge = purge_interval
        self._embedding_aggregation_mode = EmbeddingAggregationMode(embedding_aggregation_mode)
        self._writer = writer

    def append_deprecated_eavts(self, eavts: List["EAVT"]):
        entities = Entity.entities_from_deprecated_eavts(eavts).values()
        self.append_entities(entities)

    def append_entities(self, entities: List[Entity]):
        if not entities:
            logger.debug("append_entities(): No entities to append")
            return

        logger.verbose(f"append_entities(): Appending {len(entities)} entities")

        latest_frame_id = -1
        for entity in entities:
            # FIXME: we should be able to append entities without annotations
            if not entity.annotations:
                continue

            if entity.id in self._track_entities:
                self._track_entities[entity.id].append(entity)
            else:
                self._proto_tracks[entity.id].append(entity)
                # Promote proto track if it exceeds the minimum length
                if self._track_extent(self._proto_tracks[entity.id]) >= self._minimum_track_frame_length:
                    self._track_entities[entity.id] = self._proto_tracks.pop(entity.id)

            current_frame = entity.annotations[0].datum_source.frame_id
            if current_frame > latest_frame_id:
                latest_frame_id = current_frame

        if latest_frame_id >= self._next_purge:
            self.purge_proto_tracks(latest_frame_id)
            self._next_purge = latest_frame_id + self._purge_interval

    def purge_proto_tracks(self, current_frame_id: int):
        for entity_id in list(self._proto_tracks.keys()):
            last_track_frame = self._proto_tracks[entity_id][-1].annotations[0].datum_source.frame_id
            if (current_frame_id - last_track_frame) > PROTO_FRAME_SHELF_LIFE:
                del self._proto_tracks[entity_id]

    def _track_extent(self, track):
        first_track_frame_index = track[0].annotations[0].datum_source.frame_id
        last_track_frame_index = track[-1].annotations[0].datum_source.frame_id
        return last_track_frame_index - first_track_frame_index + 1

    def write(self):
        if self._writer is None:
            raise ValueError("You must pass an IEntityWriter to EntityAggregator")
        return self._writer.write(self._track_entities)



"""Data models for data source management."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional
from uuid import UUID

from pydantic import BaseModel, Field


class LocalDataSourceEntry(BaseModel):
    """Represents a single data source entry in a local file."""

    id: Optional[int] = Field(default=None, description="Cloud data source ID")
    uuid: Optional[UUID] = Field(default=None, description="Cloud data source UUID")
    name: str = Field(description="Data source name")
    source_uri: Optional[str] = Field(
        default=None, description="Source URI (e.g., rtsp://hostname:554/stream)"
    )
    mac: Optional[str] = Field(default=None, description="Device MAC address")
    serial: Optional[str] = Field(default=None, description="Device serial number")
    hostname: Optional[str] = Field(default=None, description="Device hostname")
    ip: Optional[str] = Field(default=None, description="Device IP address")
    port: Optional[str] = Field(default=None, description="Device port")
    template: Optional[str] = Field(default=None, description="Template name used")


class DataSourceMetadata(BaseModel):
    """Metadata for a data source file."""

    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class LocalDataSourceFile(BaseModel):
    """Represents a local data source file structure."""

    data_sources: List[LocalDataSourceEntry] = Field(default_factory=list)
    template: Optional[str] = Field(default=None, description="Default template for all entries")
    metadata: DataSourceMetadata = Field(default_factory=DataSourceMetadata)


class DataSourceTemplateField(BaseModel):
    """Field definition in a template."""

    required: bool = False
    type: str = "string"
    default: Optional[Any] = None


class DataSourceTemplate(BaseModel):
    """Template for creating data sources."""

    name: str
    description: str
    source_uri_pattern: str = Field(
        description="URI pattern with placeholders like {ip}, {hostname}, {port}, {serial}, {mac}, {name}"
    )
    name_pattern: Optional[str] = Field(
        default=None,
        description="Optional name pattern with placeholders like {name}, {hostname}, {ip}",
    )
    default_port: Optional[int] = None
    default_path: Optional[str] = None
    fields: Dict[str, DataSourceTemplateField] = Field(default_factory=dict)


@dataclass
class SyncResult:
    """Result of a sync or import operation."""

    matched: int = 0
    created: int = 0
    updated: int = 0
    failed: int = 0
    skipped_duplicates: int = 0
    errors: List[str] = None
    skipped_reasons: List[str] = None

    def __post_init__(self):
        if self.errors is None:
            self.errors = []
        if self.skipped_reasons is None:
            self.skipped_reasons = []


@dataclass
class MatchedDataSource:
    """A data source that has been matched between local and cloud."""

    local: LocalDataSourceEntry
    cloud: Optional[Any] = None  # DataSourceType from cloud
    match_type: Optional[Literal["id", "uuid", "name", "source_uri", "mac", "serial"]] = None
    is_matched: bool = False



import tempfile
import threading
from typing import Tuple

import aiko_services as aiko


PIPELINE_DEFINITION = """{
  "version": 0, "name": "p_test", "runtime": "python",
  "graph": ["(FrameGeneratorException)"],
  "elements": [
    { "name":   "FrameGeneratorException", "input":  [], "output": [],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_stream_lock_release" }
      }
    }
  ]
}"""


class FrameGeneratorException(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("frame_generator_exception:0")
        context.get_implementation("PipelineElement").__init__(self, context)
        self.frame_count = 0

    def start_stream(self, stream, stream_id):
        """Start stream with a frame generator that will throw an exception"""
        # Use create_frames to start the problematic _create_frames_generator thread
        test_result = self.get_parameter("test_result")[0]
        test_result["stream_lock"] = stream.lock
        self.create_frames(stream, self.failing_frame_generator, rate=10.0)
        return aiko.StreamEvent.OKAY, {}

    def failing_frame_generator(self, stream, frame_id):
        """Frame generator that throws an exception after a few frames"""
        self.frame_count += 1

        if self.frame_count <= 2:
            # Generate a couple of successful frames first
            return aiko.StreamEvent.OKAY, {"frame": self.frame_count}
        else:
            # Now throw an exception that will leave the lock unreleased
            raise RuntimeError("Simulated frame generator exception - this should cause unreleased lock!")

    def process_frame(self, stream, **kwargs) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        return aiko.StreamEvent.OKAY, {}


def test_create_frames_generator_lock_not_released_on_exception():
    """Test that the lock is released when _create_frames_generator encounters exception"""

    test_result = {"pipeline": None, "stream_lock": None}

    def run_pipeline():
        with tempfile.NamedTemporaryFile(delete=True, mode="w") as file:
            file.write(PIPELINE_DEFINITION)
            file.flush()
            pipeline_definition =  \
                aiko.PipelineImpl.parse_pipeline_definition(file.name)
            pipeline = aiko.PipelineImpl.create_pipeline(
                None, pipeline_definition, name=None, graph_path=None,
                stream_id=0, parameters={"test_result": test_result},
                frame_id=0, frame_data=None,
                grace_time=None)
            test_result["pipeline"] = pipeline
            pipeline.run(mqtt_connection_required=False)

    pipeline_thread = threading.Thread(target=run_pipeline, daemon=True)
    pipeline_thread.start()
    pipeline_thread.join(timeout=5.0)

    assert test_result["stream_lock"].in_use() is None, "Lock should be released after exception in frame generator"

    print("Test completed - Lock was released")



# Usage: File
# ~~~~~~~~~~~
# aiko_pipeline create pipelines/webcam_pipeline_0.json -s 1 -ll debug
#
# aiko_pipeline create pipelines/webcam_pipeline_0.json -s 1  \
#   -p VideoReadWebcam.path /dev/video2  # Linux
#
# aiko_pipeline create pipelines/webcam_pipeline_1.json -s 1
#
# Usage: ZMQ
# ~~~~~~~~~~
# aiko_pipeline create pipelines/image_zmq_pipeline_0.json  -s 1 -sr -gt 10
# aiko_pipeline create pipelines/image_zmq_pipeline_0.json  -s 1 -sr  \
#            -p ImageReadZMQ.data_sources zmq://0.0.0.0:6502
#
# aiko_pipeline create pipelines/webcam_zmq_pipeline_0.json -s 1 -sr \
#            -p VideoReadWebcam.rate 2.0
#            -p ImageWriteZMQ.data_targets zmq://192.168.0.1:6502
#
# Notes
# ~~~~~
# - See https://github.com/ultralytics/ultralytics/issues/3005
# pip uninstall av  # Prevent clash between AV (Ultralytics) and cv2.imshow()
# # pip install av --no-binary av
#
# To Do
# ~~~~~
# * Fix: Ensure "frame_rate" is configurable and correct for "timestamps"
#
# - Implement "VideoReadWebcam.data_batch_size" in "frame_generator()"
#
# - Implement "data_source scheme", e.g "webcam://0" or "webcam://dev/video0"
#
# - Move "import cv2" into a "source_target.py" function ?
#
# - Use "rate" (?) to control FPS (frames per seconds)
#
# - See https://docs.opencv.org/4.x/dd/d43/tutorial_py_video_display.html
#   - List available camera pathnames, e.g "/dev/video[02...]"
#   - cap.get(cv.CAP_PROP_FRAME_WIDTH),     cv.CAP_PROP_FRAME_HEIGHT
#   - cap.set(cv.CAP_PROP_FRAME_WIDTH,320), cv.CAP_PROP_FRAME_HEIGHT

import os
from typing import Tuple

import aiko_services as aiko
from aiko_services.elements.media import convert_images, open_video_capture

__all__ = ["VideoReadWebcam"]

_DEFAULT_CAMERA_PATHNAME = 0  # Linux: "/dev/video0"
_DEFAULT_FRAME_RATE = 10.0    # 10 Hz frames per second

_LOGGER = aiko.process.logger(__name__)

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "webcam_io.py: Couldn't import cv2 module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)
#   raise ModuleNotFoundError(
#       'opencv-python package not installed.  '
#       'Install aiko_services with --extras "opencv" '
#       'or install opencv-python manually to use the "webcam_io" module')

# --------------------------------------------------------------------------- #
# VideoReadWebcam is a DataSource which supports web cameras 
#
# parameter: "data_sources" is the read file path or device number
#
# Note: Only supports Streams with "data_sources" parameter

class VideoReadWebcam(aiko.DataSource):  # PipelineElement
    def __init__(self, context):
        context.set_protocol("webcam:0")
        context.call_init(self, "PipelineElement", context)

        self.path_current = None
        self.stream_started = 0
        self.video_capture = None

        self.share["color"] = True
        self.share["flip"] = "none"  # "both", "horizontal", "vertical"
        self.share["frame_id"] = -1
        self.share["path"] = _DEFAULT_CAMERA_PATHNAME
        self.ec_producer.add_handler(self._ec_producer_change_handler)

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "color":
            if isinstance(item_value, str):
                self.share["color"] = item_value.lower() == "true"

        if item_name == "path":
            if isinstance(item_value, str) and item_value.isdigit():
                item_value = int(item_value)
            if item_value != self.path_current:
                self.logger.info(f"Camera path updated: {item_value}")
                self._open_camera(item_value)

    def _close_camera(self):
        self.logger.info(f"Close camera: {self.path_current}")
        self.video_capture.release()
        self.video_capture = None
        self.path_current = None

    def _open_camera(self, path):
        if self.stream_started > 0:
            if self.video_capture:
                self._close_camera()
            if isinstance(path, str) and path.isdigit():
                path = int(path)
            self.video_capture = open_video_capture(path)
            if self.video_capture and self.video_capture.isOpened():
                self.logger.info(f"Open camera: {path}")
                self.path_current = path
                self.share["path"] = path
            else:
                self.logger.error(f"Open camera: {path} failed")
                self.video_capture = None

    def start_stream(self, stream, stream_id):
        self.stream_started += 1
        path, _ = self.get_parameter(
            "path", _DEFAULT_CAMERA_PATHNAME, self_share_priority=False)
        self._open_camera(path)
        rate, _ = self.get_parameter(
            "rate", _DEFAULT_FRAME_RATE, self_share_priority=False)
        self.create_frames(stream, self.frame_generator, rate=float(rate))
        return aiko.StreamEvent.OKAY, {}

    def frame_generator(self, stream, frame_id):
        frame_data = None
        if self.video_capture and self.video_capture.isOpened():
            status, image_bgr = self.video_capture.read()
            if status:
                if frame_id % 10 == 0:
                    self.ec_producer.update("frame_id", frame_id)
                if self.share["color"]:
                    image = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
                else:
                    image = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2GRAY)
                flip = self.share["flip"]
                if flip == "both" or flip == "horizontal":
                    image = cv2.flip(image, 1)
                if flip == "both" or flip == "vertical":
                    image = cv2.flip(image, 0)
                frame_data = {"images": [image]}
        return aiko.StreamEvent.OKAY, frame_data

    # TODO: Test this ...
    #   return aiko.StreamEvent.STOP, {"diagnostic": "Camera stopped"}

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        stream.variables["timestamps"] = [stream.frame_id * (1 / 25)]
        self.logger.debug(f"{self.my_id()}: read image")

        media_type, _ = self.get_parameter("media_type", None)
        if media_type:
            images = convert_images(images, media_type)
        return aiko.StreamEvent.OKAY, {"images": images}

    def stop_stream(self, stream, stream_id):
    #   self.create_frames_terminate(stream, self.frame_generator)
        self.stream_started -= 1
        self._close_camera()
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
